"""
This module implements the "new" binary OpenEphys format.
In this format channels are interleaved in one file.


See
https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Binary-format.html

Author: Julia Sprenger, Samuel Garcia, and Alessio Buccino
"""

import os
import json
from pathlib import Path

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class OpenEphysBinaryRawIO(BaseRawIO):
    """
    Handle several Blocks and several Segments.

    Parameters
    ----------
    dirname : str
        Path to Open Ephys directory
    load_sync_channel : bool
        If False (default) and a SYNC channel is present (e.g. Neuropixels), this is not loaded.
        If True, the SYNC channel is loaded and can be accessed in the analog signals.
    experiment_names : str or list or None
        If multiple experiments are available, this argument allows users to select one
        or more experiments. If None, all experiements are loaded as blocks.
        E.g. `experiment_names="experiment2"`, `experiment_names=["experiment1", "experiment2"]`

    Note
    ----
    For multi-experiment datasets, the streams need to be consistent across experiments.
    If this is not the case, you can select a subset of experiments with the `experiment_names`
    argument.

    # Correspondencies
    Neo          OpenEphys
    block[n-1]   experiment[n]    New device start/stop
    segment[s-1] recording[s]     New recording start/stop

    This IO handles several signal streams.
    Special event (npy) data are represented as array_annotations.
    The current implementation does not handle spiking data, this will be added upon user request

    """

    extensions = ["xml", "oebin", "txt", "dat", "npy"]
    rawmode = "one-dir"

    def __init__(self, dirname="", load_sync_channel=False, experiment_names=None):
        BaseRawIO.__init__(self)
        self.dirname = dirname
        if experiment_names is not None:
            if isinstance(experiment_names, str):
                experiment_names = [experiment_names]
        self.experiment_names = experiment_names
        self.load_sync_channel = load_sync_channel
        self.folder_structure = None
        self._use_direct_evt_timestamps = None

    def _source_name(self):
        return self.dirname

    def _parse_header(self):
        folder_structure, all_streams, nb_block, nb_segment_per_block, possible_experiments = explore_folder(
            self.dirname, self.experiment_names
        )
        check_folder_consistency(folder_structure, possible_experiments)
        self.folder_structure = folder_structure

        # all streams are consistent across blocks and segments.
        # also checks that 'continuous' and 'events' folder are present
        if "continuous" in all_streams[0][0]:
            sig_stream_names = sorted(list(all_streams[0][0]["continuous"].keys()))
        else:
            sig_stream_names = []
        if "events" in all_streams[0][0]:
            event_stream_names = sorted(list(all_streams[0][0]["events"].keys()))
        else:
            event_stream_names = []

        # first loop to reassign stream by "stream_index" instead of "stream_name"
        self._sig_streams = {}
        self._evt_streams = {}
        for block_index in range(nb_block):
            self._sig_streams[block_index] = {}
            self._evt_streams[block_index] = {}
            for seg_index in range(nb_segment_per_block[block_index]):
                self._sig_streams[block_index][seg_index] = {}
                self._evt_streams[block_index][seg_index] = {}
                for stream_index, stream_name in enumerate(sig_stream_names):
                    info_cnt = all_streams[block_index][seg_index]["continuous"][stream_name]
                    info_cnt["stream_name"] = stream_name
                    self._sig_streams[block_index][seg_index][stream_index] = info_cnt

                    # check for SYNC channel for Neuropixels streams
                    has_sync_trace = any(["SYNC" in ch["channel_name"] for ch in info_cnt["channels"]])
                    self._sig_streams[block_index][seg_index][stream_index]["has_sync_trace"] = has_sync_trace
                for i, stream_name in enumerate(event_stream_names):
                    info_evt = all_streams[block_index][seg_index]["events"][stream_name]
                    info_evt["stream_name"] = stream_name
                    self._evt_streams[block_index][seg_index][i] = info_evt

        # signals zone
        # create signals channel map: several channel per stream
        signal_channels = []
        for stream_index, stream_name in enumerate(sig_stream_names):
            # stream_index is the index in vector sytream names
            stream_id = str(stream_index)
            info = self._sig_streams[0][0][stream_index]
            new_channels = []
            for chan_info in info["channels"]:
                chan_id = chan_info["channel_name"]
                if "SYNC" in chan_id and not self.load_sync_channel:
                    continue
                if chan_info["units"] == "":
                    # in some cases for some OE version the unit is "", but the gain is to "uV"
                    units = "uV"
                else:
                    units = chan_info["units"]
                new_channels.append(
                    (
                        chan_info["channel_name"],
                        chan_id,
                        float(info["sample_rate"]),
                        info["dtype"],
                        units,
                        chan_info["bit_volts"],
                        0.0,
                        stream_id,
                    )
                )
            signal_channels.extend(new_channels)
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        signal_streams = []
        for stream_index, stream_name in enumerate(sig_stream_names):
            stream_id = str(stream_index)
            signal_streams.append((stream_name, stream_id))
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        # create memmap for signals
        for block_index in range(nb_block):
            for seg_index in range(nb_segment_per_block[block_index]):
                for stream_index, info in self._sig_streams[block_index][seg_index].items():
                    num_channels = len(info["channels"])
                    memmap_sigs = np.memmap(info["raw_filename"], info["dtype"], order="C", mode="r").reshape(
                        -1, num_channels
                    )
                    has_sync_trace = self._sig_streams[block_index][seg_index][stream_index]["has_sync_trace"]

                    # check sync channel validity (only for AP and LF)
                    if not has_sync_trace and self.load_sync_channel and "NI-DAQ" not in info["stream_name"]:
                        raise ValueError(
                            "SYNC channel is not present in the recording. " "Set load_sync_channel to False"
                        )
                    info["memmap"] = memmap_sigs

        # events zone
        # channel map: one channel one stream
        event_channels = []
        for stream_ind, stream_name in enumerate(event_stream_names):
            info = self._evt_streams[0][0][stream_ind]
            if "states" in info or "channel_states" in info:
                evt_channel_type = "epoch"
            else:
                evt_channel_type = "event"
            event_channels.append((info["channel_name"], info["channel_name"], evt_channel_type))
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # create memmap for events
        for block_index in range(nb_block):
            for seg_index in range(nb_segment_per_block[block_index]):
                for stream_index, info in self._evt_streams[block_index][seg_index].items():
                    for name in _possible_event_stream_names:
                        if name + "_npy" in info:
                            data = np.load(info[name + "_npy"], mmap_mode="r")
                            info[name] = data

                    # check that events have timestamps
                    assert "timestamps" in info, "Event stream does not have timestamps!"
                    # Updates for OpenEphys v0.6:
                    # In new vesion (>=0.6) timestamps.npy is now called sample_numbers.npy
                    # The timestamps are already in seconds, so that event times don't require scaling
                    # see https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Binary-format.html#events
                    if "sample_numbers" in info:
                        self._use_direct_evt_timestamps = True
                    else:
                        self._use_direct_evt_timestamps = False

                    # for event the neo "label" will change depending the nature
                    #  of event (ttl, text, binary)
                    # and this is transform into unicode
                    # all theses data are put in event array annotations
                    if "text" in info:
                        # text case
                        info["labels"] = info["text"].astype("U")
                    elif "metadata" in info:
                        # binary case
                        info["labels"] = info["channels"].astype("U")
                    elif "channels" in info:
                        # ttl case use channels
                        info["labels"] = info["channels"].astype("U")
                    elif "states" in info:
                        # ttl case use states
                        info["labels"] = info["states"].astype("U")
                    else:
                        raise ValueError(f"There is no possible labels for this event!")

                    # # If available, use 'states' to compute event duration
                    info["durations"] = None
                    # 'states' was introduced in OpenEphys v0.6. For previous versions, events used 'channel_states'
                    if "states" in info or "channel_states" in info:
                        states = info["channel_states"] if "channel_states" in info else info["states"]
                        if states.size > 0:
                            timestamps = info["timestamps"]
                            labels = info["labels"]
                            rising = np.where(states > 0)[0]
                            falling = np.where(states < 0)[0]

                            # infer durations
                            durations = None
                            if len(states) > 0:
                                # make sure first event is rising and last is falling
                                if states[0] < 0:
                                    falling = falling[1:]
                                if states[-1] > 0:
                                    rising = rising[:-1]

                                if len(rising) == len(falling):
                                    durations = timestamps[falling] - timestamps[rising]
                                    if not self._use_direct_evt_timestamps:
                                        timestamps = timestamps / info["sample_rate"]
                                        durations = durations / info["sample_rate"]

                            info["rising"] = rising
                            info["timestamps"] = timestamps[rising]
                            info["labels"] = labels[rising]
                            info["durations"] = durations

        # no spike read yet
        # can be implemented on user demand
        spike_channels = np.array([], dtype=_spike_channel_dtype)

        # loop for t_start/t_stop on segment browse all object
        self._t_start_segments = {}
        self._t_stop_segments = {}
        for block_index in range(nb_block):
            self._t_start_segments[block_index] = {}
            self._t_stop_segments[block_index] = {}
            for seg_index in range(nb_segment_per_block[block_index]):
                global_t_start = None
                global_t_stop = None

                # loop over signals
                for stream_index, info in self._sig_streams[block_index][seg_index].items():
                    t_start = info["t_start"]
                    dur = info["memmap"].shape[0] / float(info["sample_rate"])
                    t_stop = t_start + dur
                    if global_t_start is None or global_t_start > t_start:
                        global_t_start = t_start
                    if global_t_stop is None or global_t_stop < t_stop:
                        global_t_stop = t_stop

                # loop over events
                for stream_index, stream_name in enumerate(event_stream_names):
                    info = self._evt_streams[block_index][seg_index][stream_index]
                    if info["timestamps"].size == 0:
                        continue
                    t_start = info["timestamps"][0]
                    t_stop = info["timestamps"][-1]

                    if not self._use_direct_evt_timestamps:
                        t_start /= info["sample_rate"]
                        t_stop /= info["sample_rate"]

                    if global_t_start is None or global_t_start > t_start:
                        global_t_start = t_start
                    if global_t_stop is None or global_t_stop < t_stop:
                        global_t_stop = t_stop

                self._t_start_segments[block_index][seg_index] = global_t_start
                self._t_stop_segments[block_index][seg_index] = global_t_stop

        # main header
        self.header = {}
        self.header["nb_block"] = nb_block
        self.header["nb_segment"] = nb_segment_per_block
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # Annotate some objects from continuous files
        self._generate_minimal_annotations()
        for block_index in range(nb_block):
            bl_ann = self.raw_annotations["blocks"][block_index]
            for seg_index in range(nb_segment_per_block[block_index]):
                seg_ann = bl_ann["segments"][seg_index]

                # array annotations for signal channels
                for stream_index, stream_name in enumerate(sig_stream_names):
                    sig_ann = seg_ann["signals"][stream_index]
                    info = self._sig_streams[block_index][seg_index][stream_index]
                    has_sync_trace = self._sig_streams[block_index][seg_index][stream_index]["has_sync_trace"]

                    for k in ("identifier", "history", "source_processor_index", "recorded_processor_index"):
                        if k in info["channels"][0]:
                            values = np.array([chan_info[k] for chan_info in info["channels"]])
                            if has_sync_trace:
                                values = values[:-1]
                            sig_ann["__array_annotations__"][k] = values

                # array annotations for event channels
                # use other possible data in _possible_event_stream_names
                for stream_index, stream_name in enumerate(event_stream_names):
                    ev_ann = seg_ann["events"][stream_index]
                    info = self._evt_streams[0][0][stream_index]
                    if "rising" in info:
                        selected_indices = info["rising"]
                    else:
                        selected_indices = None
                    for k in _possible_event_stream_names:
                        if k in ("timestamps", "rising"):
                            continue
                        if k in info:
                            # split custom dtypes into separate annotations
                            if info[k].dtype.names:
                                for name in info[k].dtype.names:
                                    arr_ann = info[k][name].flatten()
                                    if selected_indices is not None:
                                        arr_ann = arr_ann[selected_indices]
                                    ev_ann["__array_annotations__"][name] = arr_ann
                            else:
                                arr_ann = info[k]
                                if selected_indices is not None:
                                    arr_ann = arr_ann[selected_indices]
                                ev_ann["__array_annotations__"][k] = arr_ann

    def _segment_t_start(self, block_index, seg_index):
        return self._t_start_segments[block_index][seg_index]

    def _segment_t_stop(self, block_index, seg_index):
        return self._t_stop_segments[block_index][seg_index]

    def _channels_to_group_id(self, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        channels = self.header["signal_channels"]
        group_ids = channels[channel_indexes]["group_id"]
        assert np.unique(group_ids).size == 1
        group_id = group_ids[0]
        return group_id

    def _get_signal_size(self, block_index, seg_index, stream_index):
        sigs = self._sig_streams[block_index][seg_index][stream_index]["memmap"]
        return sigs.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        t_start = self._sig_streams[block_index][seg_index][stream_index]["t_start"]
        return t_start

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        sigs = self._sig_streams[block_index][seg_index][stream_index]["memmap"]
        has_sync_trace = self._sig_streams[block_index][seg_index][stream_index]["has_sync_trace"]

        if not self.load_sync_channel and has_sync_trace:
            sigs = sigs[:, :-1]

        sigs = sigs[i_start:i_stop, :]
        if channel_indexes is not None:
            sigs = sigs[:, channel_indexes]
        return sigs

    def _spike_count(self, block_index, seg_index, unit_index):
        pass

    def _get_spike_timestamps(self, block_index, seg_index, unit_index, t_start, t_stop):
        pass

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        pass

    def _get_spike_raw_waveforms(self, block_index, seg_index, unit_index, t_start, t_stop):
        pass

    def _event_count(self, block_index, seg_index, event_channel_index):
        timestamps, _, _ = self._get_event_timestamps(block_index, seg_index, event_channel_index, None, None)
        return timestamps.size

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        info = self._evt_streams[block_index][seg_index][event_channel_index]
        timestamps = info["timestamps"]
        durations = info["durations"]
        labels = info["labels"]

        # slice it if needed
        if t_start is not None:
            if not self._use_direct_evt_timestamps:
                ind_start = int(t_start * info["sample_rate"])
                mask = timestamps >= ind_start
            else:
                mask = timestamps >= t_start
            timestamps = timestamps[mask]
            labels = labels[mask]
        if t_stop is not None:
            if not self._use_direct_evt_timestamps:
                ind_stop = int(t_stop * info["sample_rate"])
                mask = timestamps < ind_stop
            else:
                mask = timestamps < t_stop
            timestamps = timestamps[mask]
            labels = labels[mask]
        return timestamps, durations, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        info = self._evt_streams[0][0][event_channel_index]
        if not self._use_direct_evt_timestamps:
            event_times = event_timestamps.astype(dtype) / float(info["sample_rate"])
        else:
            event_times = event_timestamps.astype(dtype)
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        info = self._evt_streams[0][0][event_channel_index]
        if not self._use_direct_evt_timestamps:
            durations = raw_duration.astype(dtype) / float(info["sample_rate"])
        else:
            durations = raw_duration.astype(dtype)
        return durations


_possible_event_stream_names = (
    "timestamps",
    "sample_numbers",
    "channels",
    "text",
    "states",
    "full_word",
    "channel_states",
    "data_array",
    "metadata",
)


def explore_folder(dirname, experiment_names=None):
    """
    Exploring the OpenEphys folder structure, by looping through the
    folder to find recordings.

    Parameters
    ----------
    dirname (str): Root folder of the dataset

    Returns
    -------
    folder_structure: dict
        The folder_structure is dictionary that describes the Open Ephys folder.
        Dictionary structure:
        [node_name]["experiments"][exp_id]["recordings"][rec_id][stream_type][stream_information]
    all_streams: dict
        From the folder_structure, the another dictionary is reorganized with NEO-like
        indexing: block_index (experiments) and seg_index (recordings):
        Dictionary structure:
        [block_index][seg_index][stream_type][stream_information]
        where
        - node_name is the open ephys node id
        - block_index is the neo Block index
        - segment_index is the neo Segment index
        - stream_type can be 'continuous'/'events'/'spikes'
        - stream_information is a dictionary containing e.g. the sampling rate
    nb_block : int
        Number of blocks (experiments) loaded
    nb_segment_per_block : dict
        Dictionary with number of segment per block.
        Keys are block indices, values are number of segments
    possible_experiment_names : list
        List of all available experiments in the Open Ephys folder
    """
    # folder with nodes, experiments, setting files, recordings, and streams
    folder_structure = {}
    possible_experiment_names = []

    for root, dirs, files in os.walk(dirname):
        for file in files:
            if not file == "structure.oebin":
                continue
            root = Path(root)

            node_folder = root.parents[1]
            node_name = node_folder.stem
            if not node_name.startswith("Record"):
                # before version 5.x.x there was not multi Node recording
                # so no node_name
                node_name = ""

            if node_name not in folder_structure:
                folder_structure[node_name] = {}
                folder_structure[node_name]["experiments"] = {}

            # here we skip if experiment_names is not None
            experiment_folder = root.parents[0]
            experiment_name = experiment_folder.stem
            experiment_id = int(experiment_name.replace("experiment", ""))
            if experiment_name not in possible_experiment_names:
                possible_experiment_names.append(experiment_name)
            if experiment_names is not None and experiment_name not in experiment_names:
                continue
            if experiment_id not in folder_structure[node_name]["experiments"]:
                experiment = {}
                experiment["name"] = experiment_name
                if experiment_name == "experiment1":
                    settings_file = node_folder / "settings.xml"
                else:
                    settings_file = node_folder / f"settings_{experiment_id}.xml"
                experiment["settings_file"] = settings_file
                experiment["recordings"] = {}
                folder_structure[node_name]["experiments"][experiment_id] = experiment

            recording_folder = root
            recording_name = root.stem
            recording_id = int(recording_name.replace("recording", ""))
            # add recording
            recording = {}
            recording["name"] = recording_name
            recording["streams"] = {}

            # metadata
            with open(recording_folder / "structure.oebin", encoding="utf8", mode="r") as f:
                rec_structure = json.load(f)

            if (recording_folder / "continuous").exists() and len(rec_structure["continuous"]) > 0:
                recording["streams"]["continuous"] = {}
                for info in rec_structure["continuous"]:
                    # when multi Record Node the stream name also contains
                    # the node name to make it unique
                    oe_stream_name = info["folder_name"].split("/")[0]  # remove trailing slash
                    if len(node_name) > 0:
                        stream_name = node_name + "#" + oe_stream_name
                    else:
                        stream_name = oe_stream_name
                    raw_filename = recording_folder / "continuous" / info["folder_name"] / "continuous.dat"

                    # Updates for OpenEphys v0.6:
                    # In new vesion (>=0.6) timestamps.npy is now called sample_numbers.npy
                    # see https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Binary-format.html#continuous
                    sample_numbers = recording_folder / "continuous" / info["folder_name"] / "sample_numbers.npy"
                    if sample_numbers.is_file():
                        timestamp_file = sample_numbers
                    else:
                        timestamp_file = recording_folder / "continuous" / info["folder_name"] / "timestamps.npy"
                    timestamps = np.load(str(timestamp_file), mmap_mode="r")
                    timestamp0 = timestamps[0]
                    t_start = timestamp0 / info["sample_rate"]

                    # TODO for later : gap checking
                    signal_stream = info.copy()
                    signal_stream["raw_filename"] = str(raw_filename)
                    signal_stream["dtype"] = "int16"
                    signal_stream["timestamp0"] = timestamp0
                    signal_stream["t_start"] = t_start

                    recording["streams"]["continuous"][stream_name] = signal_stream

            if (root / "events").exists() and len(rec_structure["events"]) > 0:
                recording["streams"]["events"] = {}
                for info in rec_structure["events"]:
                    # when multi Record Node the stream name also contains
                    # the node name to make it unique
                    oe_stream_name = info["folder_name"].split("/")[0]  # remove trailing slash
                    if len(node_name) > 0:
                        stream_name = node_name + "#" + oe_stream_name
                    else:
                        stream_name = oe_stream_name

                    event_stream = info.copy()
                    for name in _possible_event_stream_names:
                        npy_filename = root / "events" / info["folder_name"] / f"{name}.npy"
                        if npy_filename.is_file():
                            event_stream[f"{name}_npy"] = str(npy_filename)

                    recording["streams"]["events"][stream_name] = event_stream

            folder_structure[node_name]["experiments"][experiment_id]["recordings"][recording_id] = recording

    # now create all_streams, nb_block, nb_segment_per_block
    # nested dictionary: block_index > seg_index > data_type > stream_name
    all_streams = {}
    nb_segment_per_block = {}
    record_node_names = list(folder_structure.keys())
    if len(record_node_names) == 0:
        raise ValueError(
            f"{dirname} is not a valid Open Ephys binary folder. No 'structure.oebin' "
            f"files were found in sub-folders."
        )
    recording_node = folder_structure[record_node_names[0]]

    # nb_block needs to be consistent across record nodes. Use the first one
    nb_block = len(recording_node["experiments"])

    for node_id, recording_node in folder_structure.items():
        exp_ids_sorted = sorted(list(recording_node["experiments"].keys()))
        for block_index, exp_id in enumerate(exp_ids_sorted):
            experiment = recording_node["experiments"][exp_id]
            nb_segment_per_block[block_index] = len(experiment["recordings"])
            if block_index not in all_streams:
                all_streams[block_index] = {}

            rec_ids_sorted = sorted(list(experiment["recordings"].keys()))
            for seg_index, rec_id in enumerate(rec_ids_sorted):
                recording = experiment["recordings"][rec_id]
                if seg_index not in all_streams[block_index]:
                    all_streams[block_index][seg_index] = {}
                for stream_type in recording["streams"]:
                    if stream_type not in all_streams[block_index][seg_index]:
                        all_streams[block_index][seg_index][stream_type] = {}
                    for stream_name, signal_stream in recording["streams"][stream_type].items():
                        all_streams[block_index][seg_index][stream_type][stream_name] = signal_stream

    # natural sort possible experiment names
    experiment_order = np.argsort([int(exp.replace("experiment", "")) for exp in possible_experiment_names])
    possible_experiment_names = list(np.array(possible_experiment_names)[experiment_order])

    return folder_structure, all_streams, nb_block, nb_segment_per_block, possible_experiment_names


def check_folder_consistency(folder_structure, possible_experiment_names=None):
    # check that experiment names are the same for differend record nodes
    if len(folder_structure) > 1:
        experiments = None
        for node in folder_structure.values():
            experiments_node = node["experiments"]
            if experiments is None:
                experiments = experiments_node
            experiment_names = [e["name"] for e_id, e in experiments.items()]
            assert all(
                ename["name"] in experiment_names for ename in experiments_node.values()
            ), "Inconsistent experiments across recording nodes!"

    # check that "continuous" streams are the same across multiple segments (recordings)
    record_node_names = list(folder_structure.keys())
    experiments = folder_structure[record_node_names[0]]["experiments"]
    for exp_id, experiment in experiments.items():
        segment_stream_names = None
        if len(experiment["recordings"]) > 1:
            for rec_id, recording in experiment["recordings"].items():
                stream_names = sorted(list(recording["streams"]["continuous"].keys()))
                if segment_stream_names is None:
                    segment_stream_names = stream_names
                assert segment_stream_names == stream_names, (
                    "Inconsistent continuous streams across segments! Streams for different "
                    "segments in the same experiment must be the same. Check your open ephys "
                    "folder."
                )

    # check that "continuous" streams across blocks (experiments)
    block_stream_names = None
    if len(experiments) > 1:
        for exp_id, experiment in experiments.items():
            # use 1st segment
            rec_ids = list(experiment["recordings"])
            stream_names = list(experiment["recordings"][rec_ids[0]]["streams"]["continuous"].keys())
            stream_names = sorted(stream_names)
            if block_stream_names is None:
                block_stream_names = stream_names
            assert block_stream_names == stream_names, (
                f"Inconsistent continuous streams across blocks (experiments)! Streams for "
                f"different experiments in the same folder must be the same. You can load a "
                f"subset of experiments with the 'experiment_names' argument: "
                f"{possible_experiment_names}"
            )



import quantities as pq

from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.axonrawio import AxonRawIO
from neo.core import Segment, AnalogSignal


class AxonIO(AxonRawIO, BaseFromRaw):
    """
    Class for reading data from pCLAMP and AxoScope
    files (.abf version 1 and 2), developed by Molecular device/Axon technologies.

    - abf = Axon binary file
    - atf is a text file based format from axon that could be
      read by AsciiIO (but this file is less efficient.)

    Here an important note from erikli@github for user who want to get the :
    With Axon ABF2 files, the information that you need to recapitulate the original stimulus waveform (both digital and analog) is contained in multiple places.

     - `AxonIO._axon_info['protocol']` -- things like number of samples in episode
     - `AxonIO.axon_info['section']['ADCSection']` | `AxonIO.axon_info['section']['DACSection']` -- things about the number of channels and channel properties
     - `AxonIO._axon_info['protocol']['nActiveDACChannel']` -- bitmask specifying which DACs are actually active
     - `AxonIO._axon_info['protocol']['nDigitalEnable']` -- bitmask specifying which set of Epoch timings should be used to specify the duration of digital outputs
     - `AxonIO._axon_info['dictEpochInfoPerDAC']` -- dict of dict. First index is DAC channel and second index is Epoch number (i.e. information about Epoch A in Channel 2 would be in `AxonIO._axon_info['dictEpochInfoPerDAC'][2][0]`)
     - `AxonIO._axon_info['EpochInfo']` -- list of dicts containing information about each Epoch's digital out pattern. Digital out is a bitmask with least significant bit corresponding to Digital Out 0
     - `AxonIO._axon_info['listDACInfo']` -- information about DAC name, scale factor, holding level, etc
     - `AxonIO._t_starts` -- start time of each sweep in a unified time basis
     - `AxonIO._sampling_rate`

    The current AxonIO.read_protocol() method utilizes a subset of these.
    In particular I know it doesn't consider `nDigitalEnable`, `EpochInfo`, or `nActiveDACChannel` and it doesn't account
    for different types of Epochs offered by Clampex/pClamp other than discrete steps (such as ramp, pulse train, etc and
    encoded by `nEpochType` in the EpochInfoPerDAC section). I'm currently parsing a superset of the properties used
    by read_protocol() in my analysis scripts, but that code still doesn't parse the full information and isn't in a state
    where it could be committed and I can't currently prioritize putting together all the code that would parse the full
    set of data. The `AxonIO._axon_info['EpochInfo']` section doesn't currently exist.

    """

    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename):
        AxonRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)

    def read_protocol(self):
        """
        Read the protocol waveform of the file, if present;
        function works with ABF2 only. Protocols can be reconstructed
        from the ABF1 header.

        Returns
        -------
        segments : list of segments
            Segments, one for every episode, with list of analog signls (one for every DAC).
        """
        sigs_by_segments, sig_names, sig_units = self.read_raw_protocol()
        segments = []
        for seg_index, sigs in enumerate(sigs_by_segments):
            seg = Segment(index=seg_index)
            t_start = self._t_starts[seg_index] * pq.s
            for c, sig in enumerate(sigs):
                ana_sig = AnalogSignal(
                    sig,
                    sampling_rate=self._sampling_rate * pq.Hz,
                    t_start=t_start,
                    name=sig_names[c],
                    units=sig_units[c],
                )
                seg.analogsignals.append(ana_sig)
            segments.append(seg)

        return segments



"""
This module implements :class:`ImageSequence`, a 3D array.

:class:`ImageSequence` inherits from :class:`basesignal.BaseSignal` which
derives from :class:`BaseNeo`, and from :class:`quantities.Quantity`which
in turn inherits from :class:`numpy.array`.

Inheritance from :class:`numpy.array` is explained here:
http://docs.scipy.org/doc/numpy/user/basics.subclassing.html

In brief:
* Initialization of a new object from constructor happens in :meth:`__new__`.
This is where user-specified attributes are set.

* :meth:`__array_finalize__` is called for all new objects, including those
created by slicing. This is where attributes are copied over from
the old object.
"""

import quantities as pq
import numpy as np

from neo.core.analogsignal import AnalogSignal, _get_sampling_rate
from neo.core.baseneo import BaseNeo
from neo.core.basesignal import BaseSignal
from neo.core.dataobject import DataObject


class ImageSequence(BaseSignal):
    """
    Representation of a sequence of images, as an array of three dimensions
    organized as [frame][row][column].

    Inherits from :class:`quantities.Quantity`, which in turn inherits from
    :class:`numpy.ndarray`.

    *usage*::

        >>> from neo.core import ImageSequence
        >>> import quantities as pq
        >>>
        >>> img_sequence_array = [[[column for column in range(20)]for row in range(20)]
        ...                         for frame in range(10)]
        >>> image_sequence = ImageSequence(img_sequence_array, units='V',
        ...                                sampling_rate=1 * pq.Hz,
        ...                                spatial_scale=1 * pq.micrometer)
        >>> image_sequence
        ImageSequence 10 frames with width 20 px and height 20 px; units V; datatype int64
        sampling rate: 1.0
        spatial_scale: 1.0
        >>> image_sequence.spatial_scale
        array(1.) * um

    *Required attributes/properties*:
        :image_data: (3D NumPy array, or a list of 2D arrays)
            The data itself
        :units: (quantity units)
        :sampling_rate: *or* **frame_duration** (quantity scalar) Number of
                                                samples per unit time or
                                                duration of a single image frame.
                                                If both are specified, they are
                                                checked for consistency.
        :spatial_scale: (quantity scalar) size for a pixel.
        :t_start: (quantity scalar) Time when sequence begins. Default 0.

    *Recommended attributes/properties*:
        :name: (str) A label for the dataset.
        :description: (str) Text description.
        :file_origin: (str) Filesystem path or URL of the original data file.

    *Optional attributes/properties*:
        :dtype: (numpy dtype or str) Override the dtype of the signal array.
        :copy: (bool) True by default.

    Note: Any other additional arguments are assumed to be user-specific
    metadata and stored in :attr:`annotations`.

    *Properties available on this object*:
        :sampling_rate: (quantity scalar) Number of samples per unit time.
            (1/:attr:`frame_duration`)
        :frame_duration: (quantity scalar) Duration of each image frame.
            (1/:attr:`sampling_rate`)
        :spatial_scale: Size of a pixel
        :duration: (Quantity) Sequence duration, read-only.
            (size * :attr:`frame_duration`)
        :t_stop: (quantity scalar) Time when sequence ends, read-only.
            (:attr:`t_start` + :attr:`duration`)
    """

    _parent_objects = ("Segment",)
    _parent_attrs = ("segment",)
    _quantity_attr = "image_data"
    _necessary_attrs = (
        ("image_data", pq.Quantity, 3),
        ("sampling_rate", pq.Quantity, 0),
        ("spatial_scale", pq.Quantity, 0),
        ("t_start", pq.Quantity, 0),
    )
    _recommended_attrs = BaseNeo._recommended_attrs

    def __new__(
        cls,
        image_data,
        units=pq.dimensionless,
        dtype=None,
        copy=True,
        t_start=0 * pq.s,
        spatial_scale=None,
        frame_duration=None,
        sampling_rate=None,
        name=None,
        description=None,
        file_origin=None,
        **annotations,
    ):
        """
        Constructs new :class:`ImageSequence` from data.

        This is called whenever a new class:`ImageSequence` is created from
        the constructor, but not when slicing.

        __array_finalize__ is called on the new object.
        """
        if spatial_scale is None:
            raise ValueError("spatial_scale is required")

        image_data = np.stack(image_data)
        if len(image_data.shape) != 3:
            raise ValueError("list doesn't have the correct number of dimensions")

        obj = pq.Quantity(image_data, units=units, dtype=dtype, copy=copy).view(cls)
        obj.segment = None
        # function from analogsignal.py in neo/core directory
        obj.sampling_rate = _get_sampling_rate(sampling_rate, frame_duration)
        obj.spatial_scale = spatial_scale
        if t_start is None:
            raise ValueError("t_start cannot be None")
        obj._t_start = t_start

        return obj

    def __init__(
        self,
        image_data,
        units=pq.dimensionless,
        dtype=None,
        copy=True,
        t_start=0 * pq.s,
        spatial_scale=None,
        frame_duration=None,
        sampling_rate=None,
        name=None,
        description=None,
        file_origin=None,
        **annotations,
    ):
        """
        Initializes a newly constructed :class:`ImageSequence` instance.
        """
        DataObject.__init__(self, name=name, file_origin=file_origin, description=description, **annotations)

    def __array_finalize__spec(self, obj):

        self.sampling_rate = getattr(obj, "sampling_rate", None)
        self.spatial_scale = getattr(obj, "spatial_scale", None)
        self.units = getattr(obj, "units", pq.dimensionless)
        self._t_start = getattr(obj, "_t_start", 0 * pq.s)

        return obj

    def signal_from_region(self, *region):
        """
        Method that takes 1 or multiple regionofinterest, uses the method of each region
        of interest to get the list of pixels to average.
        Return a list of :class:`AnalogSignal` for each regionofinterest
        """

        if len(region) == 0:
            raise ValueError("no regions of interest have been given")

        region_pixel = []
        for i, b in enumerate(region):
            r = region[i].pixels_in_region()
            if not r:
                raise ValueError("region " + str(i) + "is empty")
            else:
                region_pixel.append(r)
        analogsignal_list = []
        for i in region_pixel:
            data = []
            for frame in range(len(self)):
                picture_data = []
                for v in i:
                    picture_data.append(self.view(pq.Quantity)[frame][v[0]][v[1]])
                average = picture_data[0]
                for b in range(1, len(picture_data)):
                    average += picture_data[b]
                data.append((average * 1.0) / len(i))
            analogsignal_list.append(
                AnalogSignal(data, units=self.units, t_start=self.t_start, sampling_rate=self.sampling_rate)
            )

        return analogsignal_list

    def _repr_pretty_(self, pp, cycle):
        """
        Handle pretty-printing the :class:`ImageSequence`.
        """
        pp.text(
            f"{self.__class__.__name__} {self.shape[0]} frames with "
            f"width {self.shape[2]} px and height {self.shape[1]} px; "
            f"units {self.units.dimensionality.string}; datatype {self.dtype} "
        )

        def _pp(line):
            pp.breakable()
            with pp.group(indent=1):
                pp.text(line)

        for line in [f"sampling rate: {self.sampling_rate}", f"spatial_scale: {self.spatial_scale}"]:
            _pp(line)

    def _check_consistency(self, other):
        """
        Check if the attributes of another :class:`ImageSequence`
        are compatible with this one.
        """
        if isinstance(other, ImageSequence):
            for attr in ("sampling_rate", "spatial_scale", "t_start"):
                if getattr(self, attr) != getattr(other, attr):
                    raise ValueError(f"Inconsistent values of {attr}")

    # t_start attribute is handled as a property so type checking can be done
    @property
    def t_start(self):
        """
        Time when sequence begins.
        """
        return self._t_start

    @t_start.setter
    def t_start(self, start):
        """
        Setter for :attr:`t_start`
        """
        if start is None:
            raise ValueError("t_start cannot be None")
        self._t_start = start

    @property
    def duration(self):
        """
        Sequence duration

        (:attr:`size` * :attr:`frame_duration`)
        """
        return self.shape[0] / self.sampling_rate

    @property
    def t_stop(self):
        """
        Time when Sequence ends.

        (:attr:`t_start` + :attr:`duration`)
        """
        return self.t_start + self.duration

    @property
    def times(self):
        """
        The time points of each frame in the sequence

        (:attr:`t_start` + arange(:attr:`shape`)/:attr:`sampling_rate`)
        """
        return self.t_start + np.arange(self.shape[0]) / self.sampling_rate

    @property
    def frame_duration(self):
        """
        Duration of a single image frame in the sequence.

        (1/:attr:`sampling_rate`)
        """
        return 1.0 / self.sampling_rate

    @frame_duration.setter
    def frame_duration(self, duration):
        """
        Setter for :attr:`frame_duration`
        """
        if duration is None:
            raise ValueError("frame_duration cannot be None")
        elif not hasattr(duration, "units"):
            raise ValueError("frame_duration must have units")
        self.sampling_rate = 1.0 / duration



"""
Tests of neo.io.BiocamIO
"""

import unittest

from neo.io import BiocamIO
from neo.test.iotest.common_io_test import BaseTestIO


class TestBiocamIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = BiocamIO
    entities_to_download = ["biocam"]
    entities_to_test = [
        "biocam/biocam_hw3.0_fw1.6.brw",
        "biocam/biocam_hw3.0_fw1.7.0.12_raw.brw",
    ]


if __name__ == "__main__":
    unittest.main()



"""
Tests of neo.io.tdtio
"""

import unittest
from neo.io import TdtIO
from neo.test.iotest.common_io_test import BaseTestIO
import numpy as np


class TestTdtIO(
    BaseTestIO,
    unittest.TestCase,
):
    ioclass = TdtIO
    entities_to_download = ["tdt"]
    entities_to_test = [
        # test structure directory with multiple blocks
        "tdt/aep_05",
        # test single block
        "tdt/dataset_0_single_block/512ch_reconly_all-181123_B24_rest.Tdx",
        "tdt/dataset_1_single_block/ECTest-220207-135355_ECTest_B1.Tdx",
        "tdt/aep_05/Block-1/aep_05_Block-1.Tdx",
    ]

    def test_signal_group_mode(self):
        dirname = self.get_local_path("tdt/aep_05")

        # In this TDT dataset there are 3 signal streams
        nb_sigs_by_stream = [16, 1, 16]

        reader = TdtIO(dirname=dirname)
        bl = reader.read_block()
        for seg in bl.segments:
            assert len(seg.analogsignals) == 3
            i = 0
            for anasig in seg.analogsignals:
                # print(anasig.shape, anasig.sampling_rate, nb_sigs_by_stream[i])
                assert anasig.shape[1] == nb_sigs_by_stream[i]
                i += 1


if __name__ == "__main__":
    unittest.main()



"""
This module implements file reader for AlphaOmega MPX file format version 4.

This module expect default channel names from the AlphaOmega record system (RAW
###, SPK ###, LFP ###, AI ###,…).

This module reads all *.mpx files in a directory (not recursively) by default.
If you provide a list of *.lsx files only the *.mpx files referenced by those
*.lsx files will be loaded.

The specifications are mostly extracted from the "AlphaRS User Manual V1.0.1.pdf"
manual provided with the AlphaRS hardware. The specifications are described in
the chapter 6: ALPHARS FILE FORMAT. See at the end of this file for file format
blocks description.
Some informations missing from the file specifications were kindly provided by
AlphaOmega engineers.

.. note::
    Not a lot of memory optimization effort was put into this module. You should
    expect a big memory usage when loading data with this module

Author: Thomas Perret <thomas.perret@isc.cnrs.fr>
"""

import io
import mmap
import struct

from collections import defaultdict
from datetime import datetime

from pathlib import Path, PureWindowsPath

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_stream_dtype,
    _signal_channel_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)
from neo.core import NeoReadWriteError


class AlphaOmegaRawIO(BaseRawIO):
    """
    AlphaOmega MPX file format 4 reader. Handles several segments.
    A segment is a continuous recording (when recording starts/stops).
    Only files in current `dirname` are loaded, subfolders are not explored.

    Parameters
    ----------
    dirname: str | Path
        The folder from which the data will be loaded
    lsx_files: list[str] | None, default: None
        List of lsx files in `dirname` referencing mpx files to load (optional)
        If None all mpx files will be read
    prune_channels: bool, default: True
        If True removes the empty channels

    Notes
    -----
        Because channels must be gathered into coherent streams, channels names
        **must** be the default channel names in AlphaRS or Alpha LAB SNR
        software.
    """

    extensions = ["lsx", "mpx"]
    rawmode = "one-dir"

    STREAM_CHANNELS = (
        ("Accelerometer", "ACC", "ACC"),
        ("Spike Train", "SPK", "SPK"),
        ("Local Field Potential", "LFP", "LFP"),
        ("Raw Analog", "RAW", "RAW"),
        ("Analog Input", "AI", "AI"),
        ("User Defined", "UD", "UD"),
    )

    EVENT_CHANNELS = (
        ("TTL", "TTL", "TTL"),
        ("Digital Output", "DOUT", "DOUT"),
        ("User Event", "UD", "UD"),
        ("Stim Marker", "StimMarker", "StimMarker"),
        ("Digital Port", "InPort", "InPort"),
        ("Internal Detection", "Internal Detection", "Internal Detection"),
    )

    def __init__(self, dirname="", lsx_files=None, prune_channels=True):
        super().__init__(dirname=dirname)
        self.dirname = Path(dirname)

        self._lsx_files = lsx_files
        self._mpx_files = None
        self._prune_channels = prune_channels
        self._opened_files = {}
        self._ignore_unknown_datablocks = True  # internal debug property

        if self.dirname.is_dir():
            self._explore_folder()
        else:
            raise ValueError(f"{self.dirname} is not a folder")

    def _explore_folder(self):
        """
        If class was instantiated with lsx_files (list of .lsx files), load only
        the files referenced in these lsx files otherwise, load all *.mpx files
        in `dirname`.
        It does not explores the subfolders.
        """
        filenames = []
        if self._lsx_files is not None:
            for index_file in self._lsx_files:
                index_file = self.dirname / index_file
                with open(index_file, "r") as f:
                    for line in f:
                        # a line is a Microsoft Windows path. As we cannot
                        # instantiate a WindowsPath on other OS than MS
                        # Windows, we use the PureWindowsPath class
                        filename = PureWindowsPath(line.strip())
                        filename = self.dirname / filename.name
                        if not filename.is_file():
                            self.logger.warning(f"File {filename} does not exist")
                        else:
                            filenames.append(filename)
        else:
            # Load all mpx. Filter in only files in case there's a folder with
            # .mpx extension
            filenames = list(filter(lambda x: x.is_file(), self.dirname.glob("*.mpx")))
        if not filenames:
            self.logger.error(f"Found no AlphaOmega *.mpx files in {self.dirname}")
        else:
            # Sorting lexicographically should be enough to load the files in
            # correct order. This could improve slightly loading performances
            filenames.sort()
            self._mpx_files = filenames

    def _source_name(self):
        return str(self.dirname)

    def _read_file_datablocks(self, filename, prune_channels=True):
        """Read datablocks from AlphaOmega MPX file version 4.

        :param filename: the MPX filename to read datablocks from
        :type filename: Path-like object or str
        :param prune_channels: Remove references to channels and ports which
            doesn't contain any data recorded. Be careful when using this option
            with multiple-file data since it could theoretically leads to
            exception raised when data recorded in further files are merged into
            the first file pruned from these channels.
        :type prune_channels: bool
        """
        continuous_analog_channels = {}
        segmented_analog_channels = {}
        digital_channels = {}
        channel_type = {}
        stream_data_channels = {}
        ports = {}
        events = []
        unknown_datablocks = []
        with open(filename, "rb") as f:
            length, block_type = HeaderType.unpack(f.read(HeaderType.size))
            # First block is always type h and size 60
            if block_type != b"h":
                try:
                    bt = block_type.decode()
                    self.logger.error("First block must be of type h not type {bt}")
                except UnicodeDecodeError:
                    self.logger.error(
                        (f"First block must be of type h not type " f"{int.from_bytes(bt, 'little')} (int format)")
                    )
                raise Exception("First block of AlphaOmega MPX file format must be of type h")
            if not length == 60:
                self.logger.error("First block must be of size 60 (got size {length})")
                raise Exception("First block of AlphaOmega MPX file format must be of size 60")
            (
                next_block,
                version,
                hour,
                minute,
                second,
                hsecond,
                day,
                month,
                year,
                dayofweek,
                minimum_time,
                maximum_time,
                erase_count,
                map_version,
                application_name,
                resource_version,
                reserved,
            ) = SDataHeader.unpack(f.read(SDataHeader.size))
            if map_version != 4:
                self.logger.error("Only map version 4 is supported")
                raise Exception("Only AlphaOmega MPX file format 4 is supported")
            resource_version = "".join([str(b) for b in resource_version])
            try:
                resource_version = int(resource_version)
            except ValueError:
                self.logger.error(f"m_ResourceVersion should be an integer (got: {resource_version}")
            metadata = {
                "application_version": version,
                "application_name": decode_string(application_name),
                "record_date": datetime(year, month, day, hour, minute, second, 10000 * hsecond),
                "start_time": minimum_time,
                "stop_time": maximum_time,
                "erase_count": erase_count,
                "map_version": map_version,
                "resource_version": resource_version,
                "max_sample_rate": 0,
            }

            pos = 0
            while True:
                pos += length
                f.seek(pos)
                header_bytes = f.read(HeaderType.size)

                length, block_type = HeaderType.unpack(header_bytes)
                if length == 65535:
                    # The stop condition for reading MPX datablocks is base on
                    # the # length of the block: if the block has length 65535
                    # (or -1 in signed integer value) we know we have reached
                    # the end of the file
                    # We could also check that we are at the end of the file
                    # after this block
                    break

                if block_type == b"h":
                    self.logger.error("Type h block must exist only at the beginning of file")
                    raise Exception("AlphaOmega MPX file format must not have type h block after first block")

                if block_type == b"2":
                    (
                        next_block,
                        is_analog,
                        is_input,
                        channel_number,
                        *spike_color,
                    ) = Type2DataBlock.unpack(f.read(Type2DataBlock.size))
                    if is_analog and is_input:
                        (
                            mode,
                            amplitude,
                            sample_rate,
                            spike_count,
                            mode_spike,
                        ) = SDefAnalog.unpack(f.read(SDefAnalog.size))
                        mode_spike_mode = (mode_spike & 0b1111000000000000) >> 12
                        mode_spike_master = mode_spike_mode == 1
                        mode_spike_slave = mode_spike_mode == 2
                        mode_spike_linked_channel = mode_spike & 0b0000111111111111
                        mode_spike = {
                            "mode_spike_orig": mode_spike,
                            "mode_spike_master": mode_spike_master,
                            "mode_spike_slave": mode_spike_slave,
                            "mode_spike_linked_channel": mode_spike_linked_channel,
                        }
                        metadata["max_sample_rate"] = max(metadata["max_sample_rate"], sample_rate * 1000)
                        if amplitude <= 5:
                            # This is true for any logging software for map
                            # version >4 (specs say only for ALab SNR but AO
                            # engineer says it's true for any software)
                            amplitude = 1250000 / 2**15
                        if mode == 0:
                            # continuous analog channel definition block
                            if channel_number in channel_type:
                                raise NeoReadWriteError(
                                    f"channel_number {channel_number} must not be in channel_type {channel_type}"
                                )
                            channel_type[channel_number] = "continuous_analog"
                            duration, total_gain_100 = SDefContinAnalog.unpack(f.read(SDefContinAnalog.size))
                            name_length = length - 38
                            name = get_name(f, name_length)
                            if channel_number in continuous_analog_channels:
                                raise NeoReadWriteError(
                                    f"`channel_number` {channel_number} must not be in continous_analog_channels {continuous_analog_channels}"
                                )
                            continuous_analog_channels[channel_number] = {
                                "spike_color": spike_color,
                                "bit_resolution": amplitude,
                                "sample_rate": sample_rate * 1000,
                                "spike_count": spike_count,
                                "mode_spike": mode_spike,
                                "duration": duration,
                                "gain": total_gain_100 / 100,
                                "name": name,
                                "positions": defaultdict(list),
                            }
                        elif mode == 1:
                            # segmented analog channel definition block
                            if channel_number in channel_type:
                                raise NeoReadWriteError(
                                    f"`channel_number` {channel_number} must not be in channel_type {channel_type}"
                                )
                            channel_type[channel_number] = "segmented_analog"
                            (
                                pre_trig_ms,
                                post_trig_ms,
                                level_value,
                                trg_mode,
                                yes_rms,
                                total_gain_100,
                            ) = SDefLevelAnalog.unpack(f.read(SDefLevelAnalog.size))
                            name_length = length - 48
                            name = get_name(f, name_length)
                            if channel_number in segmented_analog_channels:
                                raise NeoReadWriteError(
                                    f"`channel_number` {channel_number} must be in segmented_analog_channels {segmented_analog_channels}"
                                )
                            segmented_analog_channels[channel_number] = {
                                "spike_color": spike_color,
                                "bit_resolution": amplitude,
                                "sample_rate": sample_rate * 1000,
                                "spike_count": spike_count,
                                "mode_spike": mode_spike,
                                "pre_trig_duration": pre_trig_ms / 1000,
                                "post_trig_duration": post_trig_ms / 1000,
                                "level_value": level_value,
                                "trg_mode": trg_mode,
                                "automatic_level_base_rms": yes_rms,
                                "gain": total_gain_100 / 100,
                                "name": name,
                                "positions": defaultdict(list),
                            }
                        else:
                            self.logger.error(f"Unknown type 2 analog block mode: {mode}")
                            continue
                    elif is_analog == 0 and is_input == 1:
                        # digital input channel definition
                        if channel_number in channel_type:
                            raise NeoReadWriteError(
                                f"`channel_number` {channel_number} must not be in channel_type {channel_type}"
                            )
                        channel_type[channel_number] = "digital"
                        (
                            sample_rate,
                            save_trigger,
                            duration,
                            prev_status,
                        ) = SDefDigitalInput.unpack(f.read(SDefDigitalInput.size))
                        metadata["max_sample_rate"] = max(metadata["max_sample_rate"], sample_rate * 1000)
                        if channel_number in digital_channels:
                            raise NeoReadWriteError(
                                f"`channel_number {channel_number} must not be in digital_channels {digital_channels}"
                            )
                        name_length = length - 30
                        name = get_name(f, name_length)
                        digital_channels[channel_number] = {
                            "spike_color": spike_color,
                            "sample_rate": sample_rate * 1000,
                            "save_trigger": save_trigger,
                            "duration": duration,
                            "prev_status": prev_status,
                            "name": name,
                            "samples": [],
                        }
                    else:
                        self.logger.error(f"Unknown type 2 block: analog={is_analog}, input={is_input}")
                        continue
                elif block_type == b"S":
                    # stream data definition block
                    next_block, channel_number, sample_rate = SDefStream.unpack(f.read(SDefStream.size))
                    metadata["max_sample_rate"] = max(metadata["max_sample_rate"], sample_rate * 1000)
                    if channel_number in channel_type:
                        raise NeoReadWriteError(
                            f"`channel_number` {channel_number} must not be in channel_type {channel_type}"
                        )
                    channel_type[channel_number] = "stream_data"
                    name_length = length - 18
                    name = get_name(f, name_length)
                    stream_data_channels[channel_number] = {
                        "sample_rate": sample_rate * 1000,
                        "name": name,
                    }
                elif block_type == b"b":
                    # digital input/output port definition block
                    board_number, port, sample_rate, prev_value = SDefPortX.unpack(f.read(SDefPortX.size))
                    metadata["max_sample_rate"] = max(metadata["max_sample_rate"], sample_rate * 1000)
                    if port in channel_type:
                        raise NeoReadWriteError(f"`port` {port} must not be in channel_type {channel_type}")
                    channel_type[port] = "port"
                    name_length = length - 18
                    name = get_name(f, name_length)
                    ports[port] = {
                        "board_number": board_number,
                        "sample_rate": sample_rate * 1000,
                        "prev_value": prev_value,
                        "name": name,
                        "samples": [],
                    }
                elif block_type == b"5":
                    # channel data block
                    unit_number, channel_number = SDataChannel.unpack(f.read(SDataChannel.size))
                    if channel_number not in channel_type:
                        raise ValueError(f"`channel_number` must be in channel_type {channel_type}")
                    unit_number = int.from_bytes(unit_number, "little")
                    if "analog" in channel_type[channel_number]:
                        data_length = (length - 10) / 2
                        # why do we even have this check?
                        if int(data_length) != data_length:
                            raise ValueError(f"The data length must be equal to the int of the data length")
                        data_length = int(data_length)
                        data_start = f.tell()
                        f.seek(2 * data_length, io.SEEK_CUR)
                        if channel_type[channel_number].startswith("continuous"):
                            if channel_number not in continuous_analog_channels:
                                raise NeoReadWriteError(
                                    f"The `channel_number` must be in the continuous_analog_channels {continuous_analog_channels}"
                                )
                            continuous_analog_channels[channel_number]["positions"][filename].append(
                                (
                                    SDataChannel_sample_id.unpack(f.read(SDataChannel_sample_id.size))[0],
                                    data_start,
                                    data_length,
                                )
                            )
                        elif channel_type[channel_number].startswith("segmented"):
                            if channel_number not in segmented_analog_channels:
                                raise NeoReadWriteError(
                                    f"The `channel_number` must be in the segmented_analog_channels {segmented_analog_channels}"
                                )
                            if unit_number > 0 and unit_number <= 4:
                                segmented_analog_channels[channel_number]["positions"][filename].append(
                                    (
                                        SDataChannel_sample_id.unpack(f.read(SDataChannel_sample_id.size))[0],
                                        data_start,
                                        data_length,
                                    )
                                )
                            elif unit_number == 0:
                                segmented_analog_channels[channel_number]["positions"][filename].append(
                                    (
                                        SDataChannel_sample_id.unpack(f.read(SDataChannel_sample_id.size))[0],
                                        data_start,
                                        data_length,
                                    )
                                )
                            else:
                                self.logger.error(f"Unknown unit_number={unit_number} in channel data block")
                                continue
                    elif channel_type[channel_number] == "digital":
                        if channel_number not in digital_channels:
                            raise NeoReadWriteError(
                                f"The `channel_number` must be in the digital_channels {digital_channels}"
                            )
                        sample_number, value = SDataChannelDigital.unpack(f.read(SDataChannelDigital.size))
                        digital_channels[channel_number]["samples"].append(
                            (
                                sample_number,
                                value,
                            )
                        )
                    elif channel_type[channel_number] == "port":
                        if channel_number not in ports:
                            raise NeoReadWriteError(f"The `channel_number` must be in ports {ports}")
                        # specifications says that for ports it should be "<Lh"
                        # but the data shows clearly "<HL"
                        value, sample_number = SDataChannelPort.unpack(f.read(SDataChannelPort.size))
                        ports[channel_number]["samples"].append(
                            (
                                sample_number,
                                value,
                            )
                        )
                    else:
                        self.logger.error(f"Unknown channel_type={channel_type[channel_number]} for block type 5")
                elif block_type == b"E":
                    type_event, timestamp = SAOEvent.unpack(f.read(SAOEvent.size))
                    stream_data_length = length - 8
                    events.append(
                        {
                            "timestamp": timestamp,
                            "stream_data": struct.unpack(f"<{stream_data_length}s", f.read(stream_data_length))[0],
                        }
                    )
                else:
                    if not self._ignore_unknown_datablocks:
                        try:
                            bt = block_type.decode()
                            self.logger.debug(f"Unknown block type: block length: {length}, block_type: {bt}")
                        except UnicodeDecodeError:
                            self.logger.debug(
                                (
                                    f"Unknown block type: block length: {length}, "
                                    f"block_type: {int.from_bytes(block_type, 'little')} "
                                    "(int format)"
                                )
                            )
                        unknown_datablocks.append(
                            {
                                "length": length,
                                "block_type": block_type,
                                "data": f.read(length),
                            }
                        )
        if prune_channels:
            to_remove = []
            for channel_id, channel in continuous_analog_channels.items():
                if not channel["positions"]:
                    to_remove.append(channel_id)
            for channel_id in to_remove:
                del continuous_analog_channels[channel_id]
            to_remove = []
            for channel_id, channel in segmented_analog_channels.items():
                if not channel["positions"]:
                    to_remove.append(channel_id)
            for channel_id in to_remove:
                del segmented_analog_channels[channel_id]
            to_remove = []
            for channel_id, channel in digital_channels.items():
                if not channel["samples"]:
                    to_remove.append(channel_id)
            for channel_id in to_remove:
                del digital_channels[channel_id]
            to_remove = []
            for port_id, port in ports.items():
                if not port["samples"]:
                    to_remove.append(port_id)
            for port_id in to_remove:
                del ports[port_id]
            del to_remove
        return (
            metadata,
            continuous_analog_channels,
            segmented_analog_channels,
            digital_channels,
            channel_type,
            stream_data_channels,
            ports,
            events,
            unknown_datablocks,
        )

    def _merge_segments(self, factor_period=1.5):
        """This method merge segments that are the same segment but comes from
        different files. The AlphaOmega data recording system split files with
        a configurable maximum time and size but with a global non-configurable
        maximum size of 1GB.
        Two segment are merged if they validate the following conditions:
            - start of next segment is less than `factor_period` (default 1.5)
            the system's sampling period + end of previous segment. We can't
            check with exactly 1 sampling period because timings are floats
            which are not exact period values
            - The two segment must also have the same record date
            (YEAR-MONTH-DAY). This could potentially lead to errors if
            recordings are longer than a day or run over the night
            - The next segment must have a datetime greater or equal than the
            previous segment (datetime : YEAR-MONTH-DAY-HOUR-MINUTE-SECOND)

        :param factor_period: how many sample period to consider the next segment
            is part of the previous one. This should be 1 <= `factor_period` < 2,
            defaults to 1.5
        :type factor_period: float
        """
        segments_to_merge = []
        for segment in self._segments:
            possible_same_segments = [
                s
                for s in self._segments
                if s["metadata"]["record_date"].date() == segment["metadata"]["record_date"].date()
                and s["metadata"]["record_date"] >= segment["metadata"]["record_date"]
                and 0
                <= (s["metadata"]["start_time"] - segment["metadata"]["stop_time"])
                <= factor_period / segment["metadata"]["max_sample_rate"]
                and s is not segment
            ]
            if len(possible_same_segments) not in (0, 1):
                self.logger.error(
                    f"Cannot merge segments. Found {len(possible_same_segments)} segments following segment: {segment['metadata']}"
                )
                continue
            if possible_same_segments:
                existing_merges = [s for segs in segments_to_merge for s in segs]
                if existing_merges:
                    existing_merges.sort(key=lambda x: x["metadata"]["start_time"])
                    segment = existing_merges[0]
                segments_to_merge.append((segment, possible_same_segments[0]))
        for segment, segment_to_merge in segments_to_merge:
            sample_rate = segment["metadata"]["max_sample_rate"]
            sample_rate_merge = segment_to_merge["metadata"]["max_sample_rate"]
            if sample_rate != sample_rate_merge:
                self.logger.error(
                    f"Segment to merge has sample_rate={sample_rate_merge}, expected {sample_rate}. Continuing anyway."
                )
            segment["metadata"]["stop_time"] = segment_to_merge["metadata"]["stop_time"]
            segment["metadata"]["filenames"].extend(segment_to_merge["metadata"]["filenames"])
            for stream in segment_to_merge["streams"]:
                for channel_id in segment_to_merge["streams"][stream]:
                    try:
                        channel = segment["streams"][stream][channel_id]
                    except KeyError:
                        # there can potentially have segment without stream for this
                        # channel but the segment to merge has stream for this channel
                        segment["streams"][stream][channel_id] = segment_to_merge["streams"][stream][channel_id]
                    else:
                        for f in segment_to_merge["streams"][stream][channel_id]["positions"]:
                            channel["positions"][f].extend(
                                segment_to_merge["streams"][stream][channel_id]["positions"][f]
                            )
            for channel_id in segment_to_merge["events"]:
                try:
                    channel = segment["events"][channel_id]
                except KeyError:
                    # it's possible  that the first segment doesn't record any
                    # port channel data and the port could have been pruned
                    segment["events"][channel_id] = segment_to_merge["events"][channel_id]
                else:
                    channel["samples"].extend(segment_to_merge["events"][channel_id]["samples"])
            for channel_id in segment_to_merge["spikes"]:
                try:
                    channel = segment["spikes"][channel_id]
                except KeyError:
                    # there can potentially have segment without spike for this
                    # channel but the segment to merge has spikes for this channel
                    segment["spikes"][channel_id] = segment_to_merge["spikes"][channel_id]
                else:
                    for f in segment_to_merge["spikes"][channel_id]["positions"]:
                        channel["positions"][f].extend(segment_to_merge["spikes"][channel_id]["positions"][f])
            segment["ao_events"].extend(segment_to_merge["ao_events"])
            for channel_id in segment_to_merge["stream_data"]:
                # To be honest, I have no idea what is a
                # stream_data_channels so let's just overwrite it here
                segment["stream_data"][channel_id] = segment_to_merge["stream_data"][channel_id]
            self._segments.remove(segment_to_merge)

    def _parse_header(self):
        segments = []
        for i, filename in enumerate(self._mpx_files):
            metadata, cac, sac, dc, ct, sd, p, e, ub = self._read_file_datablocks(filename, self._prune_channels)
            metadata["filenames"] = [filename]
            streams = {}
            for stream_name, channel_name_start, stream_id in self.STREAM_CHANNELS:
                channels = {i: c for i, c in cac.items() if c["name"].startswith(channel_name_start)}
                streams[stream_id] = channels
            events = dc.copy()
            events.update(p)
            segment = {
                "metadata": metadata,
                "streams": streams,
                "events": events,
                "spikes": sac,
                "ao_events": e,
                "stream_data": sd,
            }
            segments.append(segment)
        self._segments = segments
        # We merge segments after having loading all the files because they
        # could be loaded in any order
        self._merge_segments()

        signal_streams = set(
            (stream_name, stream_id)
            for segment in self._segments
            for stream in segment["streams"]
            for stream_name, _, stream_id in self.STREAM_CHANNELS
            if stream_id == stream and segment["streams"][stream]
        )
        signal_streams = list(signal_streams)
        signal_streams.sort(key=lambda x: x[1])
        signal_streams = np.array(signal_streams, dtype=_signal_stream_dtype)

        signal_channels = set(
            (
                channel["name"],
                channel_id,
                channel["sample_rate"],
                np.dtype(np.short).name,
                "uV",
                channel["gain"] / channel["bit_resolution"],
                0,
                stream,
            )
            for segment in self._segments
            for stream in segment["streams"]
            for channel_id, channel in segment["streams"][stream].items()
        )
        signal_channels = list(signal_channels)
        signal_channels.sort(key=lambda x: (x[7], x[0]))
        signal_channels = np.array(signal_channels, dtype=_signal_channel_dtype)

        spike_channels = set(
            (
                c["name"],
                i,
                "uV",
                c["gain"] / c["bit_resolution"],
                0,
                round(c["pre_trig_duration"] * c["sample_rate"]),
                c["sample_rate"],
            )
            for segment in self._segments
            for i, c in segment["spikes"].items()
        )
        spike_channels = list(spike_channels)
        spike_channels.sort(key=lambda x: x[0])
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        event_channels = set(
            (event["name"], i, "event") for segment in self._segments for i, event in segment["events"].items()
        )
        event_channels = list(event_channels)
        event_channels.sort(key=lambda x: x[1])
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [len(self._segments)]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = signal_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

        bl_ann = self.raw_annotations["blocks"][0]
        bl_ann["name"] = "Block #{}{}".format(0, " from lsx file(s) {self._lsx_files}" if self._lsx_files else "")
        bl_ann["file_origin"] = (
            "\n".join(str(self.dirname / f) for f in self._lsx_files) if self._lsx_files else bl_ann["file_origin"]
        )
        bl_ann["rec_datetime"] = self._segments[0]["metadata"]["record_date"]
        for seg_index, segment in enumerate(self._segments):
            seg_ann = bl_ann["segments"][seg_index]
            seg_ann["name"] = f"Seg #{seg_index} Block #0"
            seg_ann["file_origin"] = "\n".join(str(f) for f in self._segments[seg_index]["metadata"]["filenames"])
            seg_ann["rec_datetime"] = self._segments[seg_index]["metadata"]["record_date"]
            for c_index, c in enumerate(seg_ann["signals"]):
                c = c.copy()
                c["file_origin"] = "\n".join(
                    set(
                        str(f)
                        for channels in self._segments[seg_index]["streams"][c["stream_id"]].values()
                        for f in channels["positions"]
                    )
                )
                seg_ann["signals"][c_index] = c
            for e_index, e in enumerate(seg_ann["events"]):
                e = e.copy()
                e["file_origin"] = seg_ann["file_origin"]
                seg_ann["events"][e_index] = e

        # We open files and create mmap objects
        for filename in self._mpx_files:
            if filename not in self._opened_files:
                self._opened_files[filename] = {}
                self._opened_files[filename]["file"] = filename.open(mode="rb")
                self._opened_files[filename]["mmap"] = mmap.mmap(
                    self._opened_files[filename]["file"].fileno(),
                    0,
                    access=mmap.ACCESS_READ,
                )

    def __del__(self):
        # To be sure we close the file when object is deleted. Be aware that the
        # __del__ method is not necessarily called when interpreter exits so we
        # could still leave file opened. This is probably bad…
        for filename in self._opened_files:
            self._opened_files[filename]["mmap"].close()
            self._opened_files[filename]["file"].close()
        if hasattr(super(), "__del__"):
            super().__del__()

    def _segment_t_start(self, block_index, seg_index):
        return self._segments[seg_index]["metadata"]["start_time"]

    def _segment_t_stop(self, block_index, seg_index):
        return self._segments[seg_index]["metadata"]["stop_time"]

    def _get_signal_size(self, block_index, seg_index, stream_index):
        stream_id = self.header["signal_streams"][stream_index]["id"]
        sizes = [
            sum(sample[2] for sample_by_file in channel["positions"].values() for sample in sample_by_file)
            for channel in self._segments[seg_index]["streams"][stream_id].values()
        ]
        if not all(s == sizes[0] for s in sizes):
            raise NeoReadWriteError("The sizes of signals must be the same to get signal size")
        return sizes[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return self._segment_t_start(block_index, seg_index)

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if i_start is None:
            i_start = 0
        signal_size = self._get_signal_size(block_index, seg_index, stream_index)
        if i_stop is None or i_stop > signal_size:
            i_stop = signal_size
        stream_id = self.header["signal_streams"][stream_index]["id"]
        mask = self.header["signal_channels"]["stream_id"] == stream_id
        channel_ids = self.header["signal_channels"][mask]["id"][channel_indexes].flatten()

        # the data refers to timestamp (see docstrings) which does not start at
        # 0 but at time_start (see type H docstring)
        first_pos = []
        for channel_id in channel_ids:
            channel_id = int(channel_id)
            first_pos.append(
                min(
                    p[0]
                    for f in self._segments[seg_index]["streams"][stream_id][channel_id]["positions"].values()
                    for p in f
                )
            )

        # now we need to get all file data block indexes for the signal and
        # timestamp we want
        file_chunks = defaultdict(list)
        for i, channel_id in enumerate(channel_ids):
            channel_id = int(channel_id)
            effective_start = i_start + first_pos[i]
            effective_stop = i_stop + first_pos[i]
            for filename, positions in self._segments[seg_index]["streams"][stream_id][channel_id]["positions"].items():
                file_chunks[filename].extend(
                    [(i, *p) for p in positions if p[0] + p[2] > effective_start and p[0] < effective_stop]
                )

        # we almost surely loaded more than asked (because the blocks are not
        # contiguous) so we need to know where to cut the results
        slices_channels = []
        for channel_index, channel_id in enumerate(channel_ids):
            channel_id = int(channel_id)
            min_pos = (
                min(p[0] for f in file_chunks.values() for i, *p in f if i == channel_index) - first_pos[channel_index]
            )
            max_pos = (
                max(p[0] + p[2] for f in file_chunks.values() for i, *p in f if i == channel_index)
                - first_pos[channel_index]
            )
            slices_channels.append((min_pos, max_pos))
        min_size = min(s[0] for s in slices_channels)
        max_size = max(s[1] for s in slices_channels)
        sigs = np.ndarray((max_size - min_size, len(channel_ids)), dtype=np.short)

        for filename in file_chunks:
            # we sort by chunk position in the file because we want to optimize
            # IO access and possibly read in sequential access. This is mainly
            # true for hard drives but shouldn't hurt flash memory
            file_chunks[filename].sort(key=lambda x: x[2])
        for filename in file_chunks:
            for channel_index, chunk_index, file_position, chunk_size in file_chunks[filename]:
                sig_offset = chunk_index - first_pos[channel_index] - min_size
                sigs[sig_offset : sig_offset + chunk_size, channel_index] = np.frombuffer(
                    self._opened_files[filename]["mmap"],
                    dtype=np.short,
                    count=chunk_size,
                    offset=file_position,
                )
        return sigs[i_start - min_size : i_stop - min_size, :]

    def _spike_count(self, block_index, seg_index, spike_channel_index):
        spike_id = int(self.header["spike_channels"]["id"][spike_channel_index])
        nb_spikes = sum(len(f) for f in self._segments[seg_index]["spikes"][spike_id]["positions"].values())
        return nb_spikes

    def _get_spike_timestamps(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        if self._spike_count(block_index, seg_index, spike_channel_index):
            spike_id = int(self.header["spike_channels"]["id"][spike_channel_index])
            spikes = self._segments[seg_index]["spikes"][spike_id]
            if t_start is None:
                t_start = self._segment_t_start(block_index, seg_index)
            if t_stop is None:
                t_stop = self._segment_t_stop(block_index, seg_index)
            effective_start = t_start * spikes["sample_rate"]
            effective_stop = t_stop * spikes["sample_rate"]
            timestamps = np.array(
                [p[0] for f in spikes["positions"].values() for p in f if effective_start <= p[0] <= effective_stop]
            )
        else:
            timestamps = np.array([], dtype=np.uint32)
        return timestamps

    def _rescale_spike_timestamp(self, spike_timestamps, dtype):
        # let's hope every spike channels have the same sampling rate
        sample_rate = int(self.header["spike_channels"]["wf_sampling_rate"][0])
        spike_timestamps = spike_timestamps.astype(dtype) / sample_rate
        return spike_timestamps

    def _get_spike_raw_waveforms(self, block_index, seg_index, spike_channel_index, t_start, t_stop):
        spike_id = int(self.header["spike_channels"]["id"][spike_channel_index])
        #  nb_spikes = self._spike_count(block_index, seg_index, spike_channel_index)
        nb_spikes = self._get_spike_timestamps(block_index, seg_index, spike_channel_index, t_start, t_stop).size
        spikes = self._segments[seg_index]["spikes"][spike_id]
        spike_length = {p[2] for f in spikes["positions"].values() for p in f}
        if len(spike_length) != 1:
            raise ValueError(f"The len of `spike_length` must be 1 not {len(spike_length)}")
        spike_length = spike_length.pop()
        waveforms = np.ndarray((nb_spikes, spike_length), dtype=np.short)
        if t_start is None:
            t_start = self._segment_t_start(block_index, seg_index)
        if t_stop is None:
            t_stop = self._segment_t_stop(block_index, seg_index)
        effective_start = t_start * spikes["sample_rate"]
        effective_stop = t_stop * spikes["sample_rate"]
        i = 0
        for filename in spikes["positions"]:
            for timestamp, file_position, length in spikes["positions"][filename]:
                if effective_start <= timestamp <= effective_stop:
                    waveforms[i, :length] = np.frombuffer(
                        self._opened_files[filename]["mmap"],
                        dtype=np.short,
                        count=length,
                        offset=file_position,
                    )
                    i += 1
        waveforms.shape = nb_spikes, 1, spike_length
        return waveforms

    def _event_count(self, block_index, seg_index, event_channel_index):
        event_id = int(self.header["event_channels"]["id"][event_channel_index])
        try:
            nb_events = len(self._segments[seg_index]["events"][event_id]["samples"])
        except KeyError:
            # No event in this segment
            nb_events = 0
        return nb_events

    def _get_event_timestamps(self, block_index, seg_index, event_channel_index, t_start, t_stop):
        if self._event_count(block_index, seg_index, event_channel_index):
            event_id = int(self.header["event_channels"]["id"][event_channel_index])
            event = self._segments[seg_index]["events"][event_id]

            timestamps = np.array([s[0] for s in event["samples"]], dtype=np.uint32)
            if t_start is None:
                t_start = self._segment_t_start(block_index, seg_index)
            if t_stop is None:
                t_stop = self._segment_t_stop(block_index, seg_index)
            effective_start = t_start * event["sample_rate"]
            effective_stop = t_stop * event["sample_rate"]
            mask = (effective_start <= timestamps) & (timestamps <= effective_stop)
            timestamps = timestamps[mask]
            labels = np.array([s[1] for s in event["samples"]], dtype="U")
            labels = labels[mask]
        else:
            timestamps = np.array([], dtype=np.uint32)
            labels = np.array([], dtype="U")
        return timestamps, None, labels

    def _rescale_event_timestamp(self, event_timestamps, dtype, event_channel_index):
        event_id = int(self.header["event_channels"]["id"][event_channel_index])
        for segment in self._segments:
            if event_id in segment["events"]:
                event = segment["events"][event_id]
                break
        event_times = event_timestamps.astype(dtype) / event["sample_rate"]
        return event_times

    def _rescale_epoch_duration(self, raw_duration, dtype, event_channel_index):
        pass


def decode_string(encoded_string):
    """According to AlphaOmega engineers, all strings are NULL terminated and
    ASCII encoded"""
    return encoded_string[: encoded_string.find(b"\x00")].decode("ascii")


def get_name(f, name_length):
    """Helper function to read a string from opened binary file"""
    return decode_string(struct.unpack(f"<{name_length}s", f.read(name_length))[0])


HeaderType = struct.Struct("<Hc")
"""All datablocks start with the same common structure:
    - length (ushort): the size (in bytes) of the datablock
    - datablock_type (char): the type of datablock (described after)

There are two main datablock types:
    1. definition datablocks (types H, 2, S, B): these datablock describe metadata of
       channels and ports
    2. data datablocks (types 5, E): these datablocks contains records data of the
       previously defined channels and ports

Other datablocks exist in the data but are ignored in this implementation as per the
specification: "Any block type other than the ones described below should be
ignored."

All data is little-endian (hence the '<' in Struct calls).
"""

SDataHeader = struct.Struct("<xlhBBBBBBHBxddlB10s4sxl")
"""Type H datablock is unique and the first datablock. It specifies file metadatas:
    - alignment byte: ignore
    - next_datablock (long): offset of the next datablock from beginning of file
    - version (short): program version number
    - hour (unsigned char): start hour of data saving
    - minute (unsigned char): start minute of the data saving
    - second (unsigned char): start second of the data saving
    - hsecond (unsigned char): start 100th of seconds of the data saving
    - day (unsigned char): start day of the data saving
    - month (unsigned char): start month of the data saving
    - year (unsigned short): start year of the data saving
    - dayofweek (unsigned char): start day of week of the data saving
    - minimum_time (double): minimal acquisition time in seconds
    - maximum_time (double): maximal acquisition time in seconds
    - erase_count (long): number of erase messages in the file
    - map_version (unsigned char): MPX file format version (only 4 if supported
      by this implementation)
    - application_name (10-char string): name of the recording application.
      Should be "ARS" for AlphaRS hardware or "ALab SNR" for AlphaLab SnR hardware
    - resource_version (4-char string): C++ version used to compile recording
      software: each byte is read separately as a char and concatenated and cast
      into an int to create the original value (e.g.: b"\x00\x01\x00\x00" -> 100)
    - alignment byte: ignore
    - reserved (long): not used
"""

Type2DataBlock = struct.Struct("<xlhhhxBBB")
"""There are two (or three, depending on your interpretation) types of Type 2
datablocks:
    1. Analog channels definition
      1.a. Continuous (RAW, LFP, SPK)
      1.b. Segmented (SEG)
    2. Digital channels definition (mainly TTL)

all type 2 datablocks starts with the same structure:
    - alignment byte: ignore
    - next_datablock (long)
    - is_analog (short): 0=Digital, 1=Analog
    - is_input (short): 0=Output, 1=Input
    - channel_number: the (unique) channel number identifier from the recording software
    - alignment byte: ignore, this and the following bytes are COLORREF
      convention from Windef.h: hex value: 0x00bbggrr
    - spike_color_blue (unsigned char)
    - spike_color_green (unsigned char)
    - spike_color_red (unsigned char)
"""
SDefAnalog = struct.Struct("<hffhh")
"""
Then if is_analog and is_input:
    - mode (short): 0=Continuous, 1=(Level or Segmented)
    - amplitude (float): bit resolution. For MAP file version 4 if amplitude < 5
                         amplitude = 1_250_000/2**15
    - sample_rate (float): in kHz (or more precisely in kilosample per seconds)
    - spike_count (short): size of each data datablock (short) + timestamp (unsigned long)
    - mode_spike (2 bytes): read as hex data 0xMCCC:
        - M: 1=Master, 2=Slave
        - CCC: linked channel
        Be careful here, the first byte cover MC and the second byte the last
        part of the linked channel CC
"""
SDefContinAnalog = struct.Struct("<fh")
"""
    Then if mode is Continuous:
        - duration (float): unknown
        - total_gain_100 (short): 100 x total_gain applied to this channel
        - name (n-char string): channel name; n=length-38
"""
SDefLevelAnalog = struct.Struct("<ffhhhh")
"""
    Then if mode is Level of Segmented:
        - pre_trig_msec (float): number of milliseconds before segment trigger
        - post_trig_msec (float): number of milliseconds after segment trigger
        - level_value (short): unknown (should be the level that trigger a
          spike detection)
        - trg_mode (short): unknown (level or template mode?)
        - yes_rms (short): 1 if automatic level calculation base on RMS
        - total_gain_100 (short): see above
        - name (n-char string): channel name; n=length-48
"""
SDefDigitalInput = struct.Struct("<fhfh")
"""
Then if not is_analog and is_input
    - sample_rate (float): see above
    - save_trigger (short): unknown
    - duration (float): unknown
    - prev_status (short): not used (I guess the previously recorded value,
      could be useful for segments merged from several files)
    - name (n-char string): channel name; n=length-30
"""
"""All other combinations of is_analog and is_input are unknown (not described
in the specification and therefore not supported by this implementation)"""

SDefStream = struct.Struct("<xlhf")
"""Type S datablock: Stream data definition:
    - alignment byte: ignore
    - next_datablock (long): see above
    - channel_number (short): see above
    - sample_rate (float): see above
    - name (n-char string): channel name; n=length-18

"""

SDefPortX = struct.Struct("<xiifH")
"""Type b datablock: Digital Input/Output port definition:
    - board_number (int): not sure… maybe in case of multiple connected
      AlphaOmega setups?
    - port (int): unique port number identifier
    - sample_rate (float): in kHz (see above)
    - prev_value (ushort): not used (see above)
    - name (n-char string): port name; n=length-18
"""

SDataChannel = struct.Struct("<ch")
"""Type 5 datablock: channel data:
    - unit_number (char): for analog segmented channels: unit number; 0=Level,
      1=Unit1, 2=Unit2, 3=Unit3, 4=Unit4
    - channel_number: the previously defined channel_number (in one of the
      definition datablocks)
"""
SDataChannel_sample_id = struct.Struct("<L")
"""
Then for analog channels:
    - sample_value (n-short): array of data values
    - first_sample_number (ulong): for continuous channels: first sample number
      in the channel records. This is the timestamp (see :attr:`SAOEvent`) of
      the first sample in the data datablocks
"""
SDataChannelDigital = struct.Struct("<Lh")
"""
Then for digital channels:

    - sample_number (ulong): the sample number of the event
    - value (short): value of the event
"""
SDataChannelPort = struct.Struct("<HL")
"""
Then for digital ports:
    - value (ushort)
    - sample_number (ulong)

.. warning::
    The specification says that for port channels it should be the same as for
    digital channels but the data (and AO engineers) says otherwise. According
    to AlphaOmega engineer, this could change in a future logging software
    release and could break this reader.
"""

SAOEvent = struct.Struct("<cL")
"""Type E: stream data datablock:
    - type_event (char): event type only b"S" for now
    - timestamp (ulong): a counter initialized at 0 at hardware boot and that
      advances at the sampling rate of the system
    - stream_data (n-char): Stream Data - spec says: "Refer to SreamFormat.h or
                            use dll to decipher this stream of data"; n=length-8
"""



from neo.rawio import get_rawio
from pathlib import Path
from tempfile import TemporaryDirectory


def test_get_rawio_class():
    # use plexon io suffix for testing here
    non_existant_file = Path("non_existant_folder/non_existant_file.plx")
    non_existant_file.unlink(missing_ok=True)
    ios = get_rawio(non_existant_file)

    assert ios

    # cleanup
    non_existant_file.unlink(missing_ok=True)


def test_get_rawio_class_nonsupported_rawio():

    non_existant_file = Path("non_existant_folder/non_existant_file.fake")
    non_existant_file.unlink(missing_ok=True)
    ios = get_rawio(non_existant_file)

    assert ios is None

    # cleanup
    non_existant_file.unlink(missing_ok=True)



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.intanrawio import IntanRawIO


class IntanIO(IntanRawIO, BaseFromRaw):
    __doc__ = IntanRawIO.__doc__
    _prefered_signal_group_mode = "group-by-same-units"

    def __init__(self, filename):
        IntanRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Class for reading data from "Raw" Multi Channel System (MCS) format.
This format is NOT the native MCS format (*.mcd).
This format is a raw format with an internal binary header exported by the
"MC_DataTool binary conversion" with the option header selected.

The internal header contains sampling rate, channel names, gain and units.
Not so bad: everything that Neo needs, so this IO is without parameters.

If some MCS customers read this you should lobby to get the real specification
of the real MCS format (.mcd), then an IO module for the native MCS format
could be written instead of this ersatz.

Author: Samuel Garcia
"""

import numpy as np

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class RawMCSRawIO(BaseRawIO):
    """
    Class for reading an mcs file converted by the MC_DataToo binary converter

    Parameters
    ----------
    filename: str, default: ''
        The *.raw MCS file to be loaded

    """

    extensions = ["raw"]
    rawmode = "one-file"

    def __init__(self, filename=""):
        BaseRawIO.__init__(self)
        self.filename = filename

    def _source_name(self):
        return self.filename

    def _parse_header(self):
        self._info = info = parse_mcs_raw_header(self.filename)

        self.dtype = "uint16"
        self.sampling_rate = info["sampling_rate"]
        self.nb_channel = len(info["channel_names"])

        # one unique stream
        signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)

        self._raw_signals = np.memmap(self.filename, dtype=self.dtype, mode="r", offset=info["header_size"]).reshape(
            -1, self.nb_channel
        )

        sig_channels = []
        for c in range(self.nb_channel):
            chan_id = str(c)
            stream_id = "0"
            sig_channels.append(
                (
                    info["channel_names"][c],
                    chan_id,
                    self.sampling_rate,
                    self.dtype,
                    info["signal_units"],
                    info["signal_gain"],
                    info["signal_offset"],
                    stream_id,
                )
            )
        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        # insert some annotation at some place
        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals.shape[0] / self.sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        return self._raw_signals.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        raw_signals = self._raw_signals[slice(i_start, i_stop), channel_indexes]
        return raw_signals


def parse_mcs_raw_header(filename):
    """
    This is a from-scratch implementation, with some inspiration
    (but no code) taken from the following files:
    https://github.com/spyking-circus/spyking-circus/blob/master/circus/files/mcs_raw_binary.py
    https://github.com/jeffalstott/Multi-Channel-Systems-Import/blob/master/MCS.py
    """
    MAX_HEADER_SIZE = 5000

    with open(filename, mode="rb") as f:
        raw_header = f.read(MAX_HEADER_SIZE)

        header_size = raw_header.find(b"EOH")
        assert header_size != -1, "Error in reading raw mcs header"
        header_size = header_size + 5
        raw_header = raw_header[:header_size]
        raw_header = raw_header.replace(b"\r", b"")

        info = {}
        info["header_size"] = header_size

        def parse_line(line, key):
            if key + b" = " in line:
                v = line.replace(key, b"").replace(b" ", b"").replace(b"=", b"")
                return v

        keys = (b"Sample rate", b"ADC zero", b"ADC zero", b"El", b"Streams")

        for line in raw_header.split(b"\n"):
            for key in keys:
                v = parse_line(line, key)
                if v is None:
                    continue

                if key == b"Sample rate":
                    info["sampling_rate"] = float(v)

                elif key == b"ADC zero":
                    info["adc_zero"] = int(v)

                elif key == b"El":
                    v = v.decode("Windows-1252")
                    v = v.replace("/AD", "")
                    split_pos = 0
                    while v[split_pos] in "1234567890.":
                        split_pos += 1
                        if split_pos == len(v):
                            split_pos = None
                            break
                    assert split_pos is not None, "Impossible to find units and scaling"
                    info["signal_gain"] = float(v[:split_pos])
                    info["signal_units"] = v[split_pos:].replace("µ", "u")
                    info["signal_offset"] = -info["signal_gain"] * info["adc_zero"]

                elif key == b"Streams":
                    info["channel_names"] = v.decode("Windows-1252").split(";")

    return info



"""
Common tests for RawIOs:

It is copy/paste from neo/test/iotests/common_io_test.py

The code should be shared for common parts.


The public URL is in url_for_tests.

To deposite new testing files,  please create a account at
gin.g-node.org and upload files at NeuralEnsemble/ephy_testing_data
data repo.


"""

__test__ = False

import logging
import unittest

from neo.utils.datasets import download_dataset, get_local_testing_data_folder, default_testing_repo

from neo.test.rawiotest.tools import can_use_network

from neo.test.rawiotest import rawio_compliance as compliance

try:
    import datalad

    HAVE_DATALAD = True
except:
    HAVE_DATALAD = False

# url_for_tests = "https://portal.g-node.org/neo/" #This is the old place
repo_for_test = default_testing_repo


class BaseTestRawIO:
    """
    This class make common tests for all IOs.

    Basically download files from G-node portal.
    And test the IO is working.

    """

    # ~ __test__ = False

    # all IO test need to modify this:
    rawioclass = None  # the IOclass to be tested

    entities_to_test = []  # list of files to test compliances
    entities_to_download = []  # when files are at gin

    # allow environment to tell avoid using network
    use_network = can_use_network()

    local_test_dir = get_local_testing_data_folder()

    def setUp(self):
        """
        Set up the test fixture.  This is run for every test
        """
        self.shortname = self.rawioclass.__name__.lower().replace("rawio", "")

        if HAVE_DATALAD and self.use_network:
            for remote_path in self.entities_to_download:
                download_dataset(repo=repo_for_test, remote_path=remote_path)
        else:
            raise unittest.SkipTest("Requires datalad download of data from the web")

    def get_local_base_folder(self):
        return get_local_testing_data_folder()

    def get_local_path(self, sub_path):
        root_local_path = self.get_local_base_folder()
        local_path = root_local_path / sub_path
        # TODO later : remove the str when all IOs handle the pathlib.Path objects
        local_path = str(local_path)
        return local_path

    def test_read_all(self):
        # Read all file in self.entities_to_test
        if not HAVE_DATALAD:
            return

        for entity_name in self.entities_to_test:
            # entity_name = self.get_filename_path(entity_name)
            # local path is a folder or a file
            local_path = self.get_local_path(entity_name)

            if self.rawioclass.rawmode.endswith("-file"):
                reader = self.rawioclass(filename=local_path)
            elif self.rawioclass.rawmode.endswith("-dir"):
                reader = self.rawioclass(dirname=local_path)

            txt = reader.__repr__()
            assert "nb_block" not in txt, "Before parser_header() nb_block should be NOT known"

            reader.parse_header()

            txt = reader.__repr__()
            assert "nb_block" in txt, "After parser_header() nb_block should be known"
            # print(txt)

            #
            txt = reader._repr_annotations()
            # reader.print_annotations()

            # lanch a series of test compliance
            compliance.header_is_total(reader)
            compliance.count_element(reader)
            compliance.read_analogsignals(reader)
            compliance.read_spike_times(reader)
            compliance.read_spike_waveforms(reader)
            compliance.read_events(reader)
            compliance.has_annotations(reader)

            # basic benchmark
            level = logging.getLogger().getEffectiveLevel()
            logging.getLogger().setLevel(logging.INFO)
            compliance.benchmark_speed_read_signals(reader)
            logging.getLogger().setLevel(level)



from neo.io.basefromrawio import BaseFromRaw
from neo.rawio.maxwellrawio import MaxwellRawIO


class MaxwellIO(MaxwellRawIO, BaseFromRaw):
    mode = "file"

    def __init__(self, filename):
        MaxwellRawIO.__init__(self, filename=filename)
        BaseFromRaw.__init__(self, filename)



"""
Example for usecases.rst
"""

from itertools import cycle
import numpy as np
from quantities import ms, mV, kHz
import matplotlib.pyplot as plt
from neo import Block, Segment, ChannelView, Group, SpikeTrain, AnalogSignal

store_signals = False

block = Block(name="probe data", tetrode_ids=["Tetrode #1", "Tetrode #2"])
block.segments = [
    Segment(name="trial #1", index=0),
    Segment(name="trial #2", index=1),
    Segment(name="trial #3", index=2),
]

n_units = {"Tetrode #1": 2, "Tetrode #2": 5}

# Create a group for each neuron, annotate each group with the tetrode from which it was recorded
groups = []
counter = 0
for tetrode_id, n in n_units.items():
    groups.extend([Group(name=f"neuron #{counter + i + 1}", tetrode_id=tetrode_id) for i in range(n)])
    counter += n
block.groups.extend(groups)

iter_group = cycle(groups)

# Create dummy data, one segment at a time
for segment in block.segments:

    # create two 4-channel AnalogSignals with dummy data
    signals = {
        "Tetrode #1": AnalogSignal(np.random.rand(1000, 4) * mV, sampling_rate=10 * kHz, tetrode_id="Tetrode #1"),
        "Tetrode #2": AnalogSignal(np.random.rand(1000, 4) * mV, sampling_rate=10 * kHz, tetrode_id="Tetrode #2"),
    }
    if store_signals:
        segment.analogsignals.extend(signals.values())

    # create spike trains with dummy data
    # we will pretend the spikes have been extracted from the dummy signal
    for tetrode_id in ("Tetrode #1", "Tetrode #2"):
        for i in range(n_units[tetrode_id]):
            spiketrain = SpikeTrain(np.random.uniform(0, 100, size=30) * ms, t_stop=100 * ms)
            # assign each spiketrain to the appropriate segment
            segment.spiketrains.append(spiketrain)
            # assign each spiketrain to a given neuron
            current_group = next(iter_group)
            current_group.add(spiketrain)
            if store_signals:
                # add to the group a reference to the signal from which the spikes were obtained
                # this does not give a 1:1 correspondance between spike trains and signals,
                # for that we could use additional groups (and have groups of groups)
                current_group.add(signals[tetrode_id])


# Now plot the data

# .. by trial
plt.figure()
for seg in block.segments:
    print(f"Analyzing segment {seg.index}")
    stlist = [st - st.t_start for st in seg.spiketrains]
    plt.subplot(len(block.segments), 1, seg.index + 1)
    count, bins = np.histogram(stlist)
    plt.bar(bins[:-1], count, width=bins[1] - bins[0])
    plt.title(f"PSTH in segment {seg.index}")
plt.show()

# ..by neuron

plt.figure()
for i, group in enumerate(block.groups):
    stlist = [st - st.t_start for st in group.spiketrains]
    plt.subplot(len(block.groups), 1, i + 1)
    count, bins = np.histogram(stlist)
    plt.bar(bins[:-1], count, width=bins[1] - bins[0])
    plt.title(f"PSTH of unit {group.name}")
plt.show()

# ..by tetrode

plt.figure()
for i, tetrode_id in enumerate(block.annotations["tetrode_ids"]):
    stlist = []
    for unit in block.filter(objects=Group, tetrode_id=tetrode_id):
        stlist.extend([st - st.t_start for st in unit.spiketrains])
    plt.subplot(2, 1, i + 1)
    count, bins = np.histogram(stlist)
    plt.bar(bins[:-1], count, width=bins[1] - bins[0])
    plt.title(f"PSTH blend of tetrode {tetrode_id}")
plt.show()



"""
Tests of neo.io.neomatlabio
"""

import os
import unittest
from numpy.testing import assert_array_equal
import quantities as pq

from neo.core.analogsignal import AnalogSignal
from neo.core.irregularlysampledsignal import IrregularlySampledSignal
from neo import Block, Segment, SpikeTrain, ImageSequence, Group
from neo.test.iotest.common_io_test import BaseTestIO
from neo.test.generate_datasets import random_block

from neo.io.neomatlabio import NeoMatlabIO

try:
    import scipy.io

    HAVE_SCIPY = True
except ImportError:
    HAVE_SCIPY = False


@unittest.skipUnless(HAVE_SCIPY, "requires scipy")
class TestNeoMatlabIO(BaseTestIO, unittest.TestCase):
    ioclass = NeoMatlabIO
    files_to_test = []
    files_to_download = []

    def test_write_read_single_spike(self):
        block1 = Block(name="test_neomatlabio")
        seg = Segment("segment1")
        spiketrain1 = SpikeTrain([1] * pq.s, t_stop=10 * pq.s, sampling_rate=1 * pq.Hz)
        spiketrain1.annotate(yep="yop", yip=None)
        sig1 = AnalogSignal([4, 5, 6] * pq.A, sampling_period=1 * pq.ms)
        irrsig1 = IrregularlySampledSignal([0, 1, 2] * pq.ms, [4, 5, 6] * pq.A)
        img_sequence_array = [[[column for column in range(2)] for _ in range(2)] for _ in range(2)]
        image_sequence = ImageSequence(
            img_sequence_array, units="dimensionless", sampling_rate=1 * pq.Hz, spatial_scale=1 * pq.micrometer
        )
        block1.segments.append(seg)
        seg.spiketrains.append(spiketrain1)
        seg.analogsignals.append(sig1)
        seg.irregularlysampledsignals.append(irrsig1)
        seg.imagesequences.append(image_sequence)

        group1 = Group([spiketrain1, sig1])
        block1.groups.append(group1)

        # write block
        filename = self.get_local_path("matlabiotestfile.mat")
        io1 = self.ioclass(filename)
        io1.write_block(block1)

        # read block
        io2 = self.ioclass(filename)
        block2 = io2.read_block()

        self.assertEqual(block1.segments[0].spiketrains[0], block2.segments[0].spiketrains[0])

        assert_array_equal(block1.segments[0].analogsignals[0], block2.segments[0].analogsignals[0])

        assert_array_equal(
            block1.segments[0].irregularlysampledsignals[0].magnitude,
            block2.segments[0].irregularlysampledsignals[0].magnitude,
        )
        assert_array_equal(
            block1.segments[0].irregularlysampledsignals[0].times, block2.segments[0].irregularlysampledsignals[0].times
        )

        assert_array_equal(block1.segments[0].imagesequences[0], block2.segments[0].imagesequences[0])

        # test annotations
        spiketrain2 = block2.segments[0].spiketrains[0]
        assert spiketrain2.annotations["yep"] == "yop"
        assert spiketrain2.annotations["yip"] is None

        # test group retrieval
        group2 = block2.groups[0]
        assert_array_equal(group1.analogsignals[0], group2.analogsignals[0])

    def test_write_read_random_blocks(self):
        for i in range(10):
            # generate random block
            block1 = random_block()

            # write block to file
            filename_orig = self.get_local_path(f"matlabio_randomtest_orig_{i}.mat")
            io1 = self.ioclass(filename_orig)
            io1.write_block(block1)

            # read block
            io2 = self.ioclass(filename_orig)
            block2 = io2.read_block()

            filename_roundtripped = self.get_local_path(f"matlabio_randomtest_roundtrip_{i}.mat")
            io3 = self.ioclass(filename_roundtripped)
            io3.write_block(block2)

            # the actual contents will differ since we're using Python object id as identifiers
            # but at least the file size should be the same
            assert os.stat(filename_orig).st_size == os.stat(filename_roundtripped).st_size


if __name__ == "__main__":
    unittest.main()



"""
Reading from neuroscope format files.
Ref: http://neuroscope.sourceforge.net/

It is an old format from Buzsaki's lab

Some old open datasets from spike sorting
are still using this format.

This only the signals.
This should be done (but maybe never will):
  * SpikeTrain file   '.clu'  '.res'
  * Event  '.ext.evt'  or '.evt.ext'

Author: Samuel Garcia

"""

from pathlib import Path

import numpy as np
from xml.etree import ElementTree

from .baserawio import (
    BaseRawIO,
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
)


class NeuroScopeRawIO(BaseRawIO):
    extensions = ["xml", "dat", "lfp", "eeg"]
    rawmode = "one-file"

    def __init__(self, filename, binary_file=None):
        """raw reader for Neuroscope

        Parameters
        ----------
        filename : str, Path
            Usually the path of an xml file
        binary_file : str | Path | None, default: None
            The binary data file
            Supported formats: ['.dat', '.lfp', '.eeg']

        Neuroscope format is composed of two files: a xml file with metadata and a binary file
        in either .dat, .lfp or .eeg format.

        Notes
        -----
        For backwards compatibility, we offer three ways of initializing the reader.

        Cases:
        filename provided with .xml extension:
            - If binary_file is provided, it is used as the data file.
            - If binary_file is not provided, it tries to find a binary file with the same name and the
            supported extensions (.dat, .lfp, .eeg) in that order.
        filename provided with empty extension:
            - If binary_file is provided, it is used as the data file.
            - If binary_file is not provided, it tries to find a binary file with the same name and the
            supported extensions (.dat, .lfp, .eeg) in that order.
        filename provided with a supported data extension (.dat, .lfp, .eeg):
            - It assumes that the XML file has the same name and a .xml extension.
        """
        BaseRawIO.__init__(self)
        self.filename = filename
        self.binary_file = binary_file

    def _source_name(self):
        return Path(self.filename).stem

    def _parse_header(self):
        # Load the right paths to xml and data
        self._resolve_xml_and_data_paths()

        # Parse XML-file
        tree = ElementTree.parse(self.xml_file_path)
        root = tree.getroot()
        acq = root.find("acquisitionSystem")
        nbits = int(acq.find("nBits").text)
        nb_channel = int(acq.find("nChannels").text)
        self._sampling_rate = float(acq.find("samplingRate").text)
        voltage_range = float(acq.find("voltageRange").text)
        # offset = int(acq.find('offset').text)
        amplification = float(acq.find("amplification").text)

        # find groups for channels
        channel_group = {}
        for grp_index, xml_chx in enumerate(root.find("anatomicalDescription").find("channelGroups").findall("group")):
            for xml_rc in xml_chx:
                channel_group[int(xml_rc.text)] = grp_index

        sig_dtype = "int16"
        # scale to convert sample values to voltage in mV = range of recording in volts *
        #  1000 mV/V /(number of bits * amplification)
        if nbits == 16:
            gain = voltage_range * 1000 / (2**nbits) / amplification
            # ~ elif nbits==32:
            # Not sure if it is int or float
            # ~ dt = 'int32'
            # ~ gain  = voltage_range/2**32/amplification
        else:
            raise (NotImplementedError)

        # Extract signal from the data file
        self._raw_signals = np.memmap(self.data_file_path, dtype=sig_dtype, mode="r", offset=0).reshape(-1, nb_channel)

        # one unique stream
        signal_streams = np.array([("Signals", "0")], dtype=_signal_stream_dtype)

        # signals
        sig_channels = []
        for c in range(nb_channel):
            name = f"ch{c}grp{channel_group.get(c, 'none')}"
            chan_id = str(c)
            units = "mV"
            offset = 0.0
            stream_id = "0"
            sig_channels.append((name, chan_id, self._sampling_rate, sig_dtype, units, gain, offset, stream_id))
        sig_channels = np.array(sig_channels, dtype=_signal_channel_dtype)

        # No events
        event_channels = []
        event_channels = np.array(event_channels, dtype=_event_channel_dtype)

        # No spikes
        spike_channels = []
        spike_channels = np.array(spike_channels, dtype=_spike_channel_dtype)

        # fille into header dict
        self.header = {}
        self.header["nb_block"] = 1
        self.header["nb_segment"] = [1]
        self.header["signal_streams"] = signal_streams
        self.header["signal_channels"] = sig_channels
        self.header["spike_channels"] = spike_channels
        self.header["event_channels"] = event_channels

        self._generate_minimal_annotations()

    def _segment_t_start(self, block_index, seg_index):
        return 0.0

    def _segment_t_stop(self, block_index, seg_index):
        t_stop = self._raw_signals.shape[0] / self._sampling_rate
        return t_stop

    def _get_signal_size(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return self._raw_signals.shape[0]

    def _get_signal_t_start(self, block_index, seg_index, stream_index):
        if stream_index != 0:
            raise ValueError("`stream_index` must be 0")
        return 0.0

    def _get_analogsignal_chunk(self, block_index, seg_index, i_start, i_stop, stream_index, channel_indexes):
        if channel_indexes is None:
            channel_indexes = slice(None)
        raw_signals = self._raw_signals[slice(i_start, i_stop), channel_indexes]
        return raw_signals

    def _resolve_xml_and_data_paths(self):
        """
        Resolves XML and data paths from the provided filename and binary_file attributes.

        See the __init__ of the class for more a description of the conditions.

        Using these conditions these function updates the self.xml_file_path and self.data_file_path attributes.

        """

        supported_extensions = [".dat", ".lfp", ".eeg"]
        self.filename = Path(self.filename)
        self.binary_file = Path(self.binary_file) if self.binary_file is not None else None

        if self.filename.suffix == ".xml":
            xml_file_path = self.filename
            data_file_path = self.binary_file
        elif self.filename.suffix == "":
            xml_file_path = self.filename.with_suffix(".xml")
            data_file_path = self.binary_file
        elif self.filename.suffix in supported_extensions:
            xml_file_path = self.filename.with_suffix(".xml")
            data_file_path = self.filename
        else:
            raise KeyError(
                f"Format {self.filename.suffix} not supported, filename format should be {supported_extensions} or .xml"
            )

        if data_file_path is None:
            possible_file_paths = (xml_file_path.with_suffix(extension) for extension in supported_extensions)
            data_file_path = next((path for path in possible_file_paths if path.is_file()), None)
            if data_file_path is None:
                raise FileNotFoundError(
                    f"data binary not found for file {xml_file_path.stem} with supported extensions: {supported_extensions}"
                )

        if not xml_file_path.is_file():
            raise FileNotFoundError(f"XML file not found at the expected location {xml_file_path}")
        if not data_file_path.is_file():
            raise FileNotFoundError(f"Binary file not found at the expected location {data_file_path}")

        self.xml_file_path = xml_file_path
        self.data_file_path = data_file_path



"""
Here a list for testing neo.rawio API compliance.
This is called automatically by `BaseTestRawIO`

All rules are listed as function so it should be easier to:
  * identify the rawio API
  * debug
  * discuss rules

"""

import time

if not hasattr(time, "perf_counter"):
    time.perf_counter = time.time
import logging

import numpy as np

from neo.rawio.baserawio import (
    _signal_channel_dtype,
    _signal_stream_dtype,
    _spike_channel_dtype,
    _event_channel_dtype,
    _common_sig_characteristics,
)


def print_class(reader):
    return reader.__class__.__name__


def header_is_total(reader):
    """
    Test if hedaer contains:
      * 'nb_block'
      * 'nb_segment'
      * 'signal_streams'
      * 'signal_channels'
      * 'spike_channels'
      * 'event_channels'

    """
    h = reader.header

    assert "nb_block" in h, "`nb_block`missing in header"
    assert "nb_segment" in h, "`nb_segment`missing in header"
    assert len(h["nb_segment"]) == h["nb_block"]

    assert "signal_streams" in h, "signal_streams missing in header"
    if h["signal_streams"] is not None:
        dt = h["signal_streams"].dtype
        for k, _ in _signal_stream_dtype:
            assert k in dt.fields, f"{k} not in signal_streams.dtype"

    assert "signal_channels" in h, "signal_channels missing in header"
    if h["signal_channels"] is not None:
        dt = h["signal_channels"].dtype
        for k, _ in _signal_channel_dtype:
            assert k in dt.fields, f"{k} not in signal_channels.dtype"

    assert "spike_channels" in h, "spike_channels missing in header"
    if h["spike_channels"] is not None:
        dt = h["spike_channels"].dtype
        for k, _ in _spike_channel_dtype:
            assert k in dt.fields, f"{k} not in spike_channels.dtype"

    assert "event_channels" in h, "event_channels missing in header"
    if h["event_channels"] is not None:
        dt = h["event_channels"].dtype
        for k, _ in _event_channel_dtype:
            assert k in dt.fields, f"{k} not in event_channels.dtype"


def count_element(reader):
    """
    Count block/segment/signals/spike/events

    """

    nb_stream = reader.signal_streams_count()
    for stream_index in range(nb_stream):
        nb_chan = reader.signal_channels_count(stream_index)
    nb_unit = reader.spike_channels_count()
    nb_event_channel = reader.event_channels_count()

    nb_block = reader.block_count()
    assert nb_block > 0, f"{print_class(reader)} have {nb_block} block"

    for block_index in range(nb_block):
        nb_seg = reader.segment_count(block_index)

        for seg_index in range(nb_seg):
            t_start = reader.segment_t_start(block_index=block_index, seg_index=seg_index)
            t_stop = reader.segment_t_stop(block_index=block_index, seg_index=seg_index)
            assert t_stop > t_start

            for stream_index in range(nb_stream):
                sig_size = reader.get_signal_size(block_index, seg_index, stream_index=stream_index)

                for spike_channel_index in range(nb_unit):
                    nb_spike = reader.spike_count(
                        block_index=block_index, seg_index=seg_index, spike_channel_index=spike_channel_index
                    )

                for event_channel_index in range(nb_event_channel):
                    nb_event = reader.event_count(
                        block_index=block_index, seg_index=seg_index, event_channel_index=event_channel_index
                    )


def iter_over_sig_chunks(reader, stream_index, channel_indexes, chunksize=1024):
    nb_block = reader.block_count()

    # read all chunk in RAW data
    for block_index in range(nb_block):
        nb_seg = reader.segment_count(block_index)
        for seg_index in range(nb_seg):
            sig_size = reader.get_signal_size(block_index, seg_index, stream_index)

            nb = sig_size // chunksize + 1
            for i in range(nb):
                i_start = i * chunksize
                i_stop = min((i + 1) * chunksize, sig_size)
                raw_chunk = reader.get_analogsignal_chunk(
                    block_index=block_index,
                    seg_index=seg_index,
                    i_start=i_start,
                    i_stop=i_stop,
                    stream_index=stream_index,
                    channel_indexes=channel_indexes,
                )
                yield raw_chunk


def read_analogsignals(reader):
    """
    Read and convert some signals chunks.

    Test special case when signal_channels do not have same sampling_rate.
    AKA _need_chan_index_check
    """
    nb_stream = reader.signal_streams_count()
    if nb_stream == 0:
        return

    for stream_index in range(nb_stream):
        sr = reader.get_signal_sampling_rate(stream_index=stream_index)
        assert type(sr) == float, f"Type of sampling is {type(sr)} should float"

        # make other test on the first chunk of first block first block
        block_index = 0
        seg_index = 0

        sig_size = reader.get_signal_size(block_index, seg_index, stream_index)

        # read all chunk for all channel all block all segment
        channel_indexes = None
        for raw_chunk in iter_over_sig_chunks(reader, stream_index, channel_indexes, chunksize=1024):
            assert raw_chunk.ndim == 2

        i_start = 0
        sig_size = reader.get_signal_size(block_index, seg_index, stream_index)
        i_stop = min(1024, sig_size)

        nb_chan = reader.signal_channels_count(stream_index)
        channel_indexes = np.arange(nb_chan, dtype=int)

        signal_channels = reader.header["signal_channels"]
        stream_id = reader.header["signal_streams"][stream_index]["id"]
        mask = signal_channels["stream_id"] == stream_id
        channel_names = signal_channels["name"][mask]
        channel_ids = signal_channels["id"][mask]

        # acces by channel inde/ids/names should give the same chunk
        channel_indexes2 = channel_indexes[::2]
        channel_names2 = channel_names[::2]
        channel_ids2 = channel_ids[::2]

        # slice by index
        raw_chunk0 = reader.get_analogsignal_chunk(
            block_index=block_index,
            seg_index=seg_index,
            i_start=i_start,
            i_stop=i_stop,
            stream_index=stream_index,
            channel_indexes=channel_indexes2,
        )
        assert raw_chunk0.ndim == 2
        assert raw_chunk0.shape[0] == i_stop
        assert raw_chunk0.shape[1] == len(channel_indexes2)

        # slice by ids
        raw_chunk2 = reader.get_analogsignal_chunk(
            block_index=block_index,
            seg_index=seg_index,
            i_start=i_start,
            i_stop=i_stop,
            stream_index=stream_index,
            channel_ids=channel_ids2,
        )
        np.testing.assert_array_equal(raw_chunk0, raw_chunk2)

        # channel names are not always unique inside a stream
        unique_chan_name = np.unique(channel_names).size == channel_names.size
        if unique_chan_name:
            raw_chunk1 = reader.get_analogsignal_chunk(
                block_index=block_index,
                seg_index=seg_index,
                i_start=i_start,
                i_stop=i_stop,
                stream_index=stream_index,
                channel_names=channel_names2,
            )
            np.testing.assert_array_equal(raw_chunk0, raw_chunk1)

        # test prefer_slice=True/False
        if nb_chan >= 3:
            for prefer_slice in (True, False):
                raw_chunk3 = reader.get_analogsignal_chunk(
                    block_index=block_index,
                    seg_index=seg_index,
                    i_start=i_start,
                    i_stop=i_stop,
                    stream_index=stream_index,
                    channel_indexes=[1, 2],
                )

        # convert to float32/float64
        for dt in ("float32", "float64"):
            float_chunk0 = reader.rescale_signal_raw_to_float(
                raw_chunk0, dtype=dt, stream_index=stream_index, channel_indexes=channel_indexes2
            )
            float_chunk2 = reader.rescale_signal_raw_to_float(
                raw_chunk2, dtype=dt, stream_index=stream_index, channel_ids=channel_ids2
            )
            if unique_chan_name:
                float_chunk1 = reader.rescale_signal_raw_to_float(
                    raw_chunk1, dtype=dt, stream_index=stream_index, channel_names=channel_names2
                )

            assert float_chunk0.dtype == dt
            np.testing.assert_array_equal(float_chunk0, float_chunk2)
            if unique_chan_name:
                np.testing.assert_array_equal(float_chunk0, float_chunk1)

        # read 500ms with several chunksize
        sr = reader.get_signal_sampling_rate(stream_index=stream_index)
        lenght_to_read = int(0.5 * sr)
        if lenght_to_read < sig_size:
            ref_raw_sigs = reader.get_analogsignal_chunk(
                block_index=block_index,
                seg_index=seg_index,
                i_start=0,
                i_stop=lenght_to_read,
                stream_index=stream_index,
                channel_indexes=channel_indexes,
            )
            for chunksize in (511, 512, 513, 1023, 1024, 1025):
                i_start = 0
                chunks = []
                while i_start < lenght_to_read:
                    i_stop = min(i_start + chunksize, lenght_to_read)
                    raw_chunk = reader.get_analogsignal_chunk(
                        block_index=block_index,
                        seg_index=seg_index,
                        i_start=i_start,
                        i_stop=i_stop,
                        stream_index=stream_index,
                        channel_indexes=channel_indexes,
                    )
                    chunks.append(raw_chunk)
                    i_start += chunksize
                chunk_raw_sigs = np.concatenate(chunks, axis=0)

                np.testing.assert_array_equal(ref_raw_sigs, chunk_raw_sigs)


def benchmark_speed_read_signals(reader):
    """
    A very basic speed measurement that read all signal
    in a file.
    """

    nb_stream = reader.signal_streams_count()
    if nb_stream == 0:
        return

    for stream_index in range(nb_stream):

        nb_samples = 0
        channel_indexes = None
        nb_chan = reader.signal_channels_count(stream_index)

        t0 = time.perf_counter()
        for raw_chunk in iter_over_sig_chunks(reader, stream_index, channel_indexes, chunksize=1024):
            nb_samples += raw_chunk.shape[0]
        t1 = time.perf_counter()
        if t0 != t1:
            speed = (nb_samples * nb_chan) / (t1 - t0) / 1e6
        else:
            speed = np.inf
        logging.info(
            f"{print_class(reader)} read ({nb_chan}channels x {nb_samples}samples)"
            f"in {t1 - t0:0.3f} s so speed {speed:0.3f} MSPS from {reader.source_name()}"
        )


def read_spike_times(reader):
    """
    Read and convert all spike times.
    """

    nb_block = reader.block_count()
    nb_unit = reader.spike_channels_count()

    for block_index in range(nb_block):
        nb_seg = reader.segment_count(block_index)
        for seg_index in range(nb_seg):
            for spike_channel_index in range(nb_unit):
                nb_spike = reader.spike_count(
                    block_index=block_index, seg_index=seg_index, spike_channel_index=spike_channel_index
                )
                if nb_spike == 0:
                    continue

                spike_timestamp = reader.get_spike_timestamps(
                    block_index=block_index,
                    seg_index=seg_index,
                    spike_channel_index=spike_channel_index,
                    t_start=None,
                    t_stop=None,
                )
                assert spike_timestamp.shape[0] == nb_spike, f"nb_spike {spike_timestamp.shape[0]} != {nb_spike}"

                spike_times = reader.rescale_spike_timestamp(spike_timestamp, "float64")
                assert spike_times.dtype == "float64"

                if spike_times.size > 3:
                    # load only one spike by forcing limits
                    t_start = spike_times[1] - 0.001
                    t_stop = spike_times[1] + 0.001

                    spike_timestamp2 = reader.get_spike_timestamps(
                        block_index=block_index,
                        seg_index=seg_index,
                        spike_channel_index=spike_channel_index,
                        t_start=t_start,
                        t_stop=t_stop,
                    )
                    assert spike_timestamp2.shape[0] == 1

                    spike_times2 = reader.rescale_spike_timestamp(spike_timestamp2, "float64")
                    assert spike_times2[0] == spike_times[1]


def read_spike_waveforms(reader):
    """
    Read and convert some all waveforms.
    """
    nb_block = reader.block_count()
    nb_unit = reader.spike_channels_count()

    for block_index in range(nb_block):
        nb_seg = reader.segment_count(block_index)
        for seg_index in range(nb_seg):
            for spike_channel_index in range(nb_unit):
                nb_spike = reader.spike_count(
                    block_index=block_index, seg_index=seg_index, spike_channel_index=spike_channel_index
                )
                if nb_spike == 0:
                    continue

                raw_waveforms = reader.get_spike_raw_waveforms(
                    block_index=block_index,
                    seg_index=seg_index,
                    spike_channel_index=spike_channel_index,
                    t_start=None,
                    t_stop=None,
                )
                if raw_waveforms is None:
                    continue
                assert raw_waveforms.shape[0] == nb_spike
                assert raw_waveforms.ndim == 3

                for dt in ("float32", "float64"):
                    float_waveforms = reader.rescale_waveforms_to_float(
                        raw_waveforms, dtype=dt, spike_channel_index=spike_channel_index
                    )
                    assert float_waveforms.dtype == dt
                    assert float_waveforms.shape == raw_waveforms.shape


def read_events(reader):
    """
    Read and convert some event or epoch.
    """
    nb_block = reader.block_count()
    nb_event_channel = reader.event_channels_count()

    for block_index in range(nb_block):
        nb_seg = reader.segment_count(block_index)
        for seg_index in range(nb_seg):
            for ev_chan in range(nb_event_channel):
                nb_event = reader.event_count(block_index=block_index, seg_index=seg_index, event_channel_index=ev_chan)
                if nb_event == 0:
                    continue

                ev_timestamps, ev_durations, ev_labels = reader.get_event_timestamps(
                    block_index=block_index, seg_index=seg_index, event_channel_index=ev_chan
                )
                assert ev_timestamps.shape[0] == nb_event, f"Wrong shape {ev_timestamps.shape[0]}, {nb_event}"
                if ev_durations is not None:
                    assert ev_durations.shape[0] == nb_event
                assert ev_labels.shape[0] == nb_event

                ev_times = reader.rescale_event_timestamp(ev_timestamps, dtype="float64")
                assert ev_times.dtype == "float64"


def has_annotations(reader):
    assert hasattr(reader, "raw_annotations"), "raw_annotation are not set"



"""
This module defines :class:`BaseNeo`, the abstract base class
used by all :module:`neo.core` classes.
"""

from copy import deepcopy
from datetime import datetime, date, time, timedelta
from decimal import Decimal
import logging
from numbers import Number

import numpy as np

ALLOWED_ANNOTATION_TYPES = (
    int,
    float,
    complex,
    str,
    bytes,
    type(None),
    datetime,
    date,
    time,
    timedelta,
    Number,
    Decimal,
    np.number,
    np.bool_,
)

logger = logging.getLogger("Neo")


class MergeError(Exception):
    pass


class NeoReadWriteError(IOError):
    """
    This is the main neo-specific error that has to deal with
    ephys data that does not follow the requirements for the Neo
    object hierarchy.

    It should be used for:
        1) When an IO does not support a specific read/write functionality
        2) A file format does not support the structural requirements
            * Different sampling rates among streams
            * Different expectations for a file format (could also be
            a NotImplementedError depending on circumstances)

    It should NOT be used when other errors more accurately describe
    the problem:
        1) ValueError: for incorrect values (like t_start, t_stop)
        2) TypeError: use of an inappropriate type for an argument
        3) FileNotFoundError: for use when a file is not found
    """

    pass


def _check_annotations(value):
    """
    Recursively check that value is either of a "simple" type (number, string,
    date/time) or is a (possibly nested) dict, list or numpy array containing
    only simple types.
    """
    if isinstance(value, np.ndarray):
        if not issubclass(value.dtype.type, ALLOWED_ANNOTATION_TYPES):
            raise ValueError(f"Invalid annotation. NumPy arrays with dtype {value.dtype.type}" f"are not allowed")
    elif isinstance(value, dict):
        for key, element in value.items():
            if not isinstance(key, str):
                raise TypeError(f"Annotations keys must be strings not of type {type(key)}")
            _check_annotations(element)
    elif isinstance(value, (list, tuple)):
        for element in value:
            _check_annotations(element)
    elif not isinstance(value, ALLOWED_ANNOTATION_TYPES):
        raise ValueError(f"Invalid annotation. Annotations of type {type(value)} are not" f"allowed")


def merge_annotation(a, b):
    """
    First attempt at a policy for merging annotations (intended for use with
    parallel computations using MPI). This policy needs to be discussed
    further, or we could allow the user to specify a policy.

    Current policy:
        For arrays or lists: concatenate
        For dicts: merge recursively
        For strings: concatenate with ';'
        Otherwise: fail if the annotations are not equal
    """
    if type(a) != type(b):
        raise TypeError(f"type({a}) {type(a)} != type({b}) {type(b)}")
    if isinstance(a, dict):
        return merge_annotations(a, b)
    elif isinstance(a, np.ndarray):  # concatenate b to a
        return np.append(a, b)
    elif isinstance(a, list):  # concatenate b to a
        return a + b
    elif isinstance(a, str):
        if a == b:
            return a
        else:
            return a + ";" + b
    else:
        if a != b:
            raise ValueError(f"{a} != {b}")
        return a


def merge_annotations(A, *Bs):
    """
    Merge two sets of annotations.

    Merging follows these rules:
    All keys that are in A or B, but not both, are kept.
    For keys that are present in both:
        For arrays or lists: concatenate
        For dicts: merge recursively
        For strings: concatenate with ';'
        Otherwise: warn if the annotations are not equal
    """
    merged = A.copy()
    for B in Bs:
        for name in B:
            if name not in merged:
                merged[name] = B[name]
            else:
                try:
                    merged[name] = merge_annotation(merged[name], B[name])
                except BaseException as exc:
                    # exc.args += ('key %s' % name,)
                    # raise
                    merged[name] = "MERGE CONFLICT"  # temporary hack
    logger.debug(f"Merging annotations: A={A} Bs={Bs} merged={merged}")
    return merged


def intersect_annotations(A, B):
    """
    Identify common entries in dictionaries A and B
    and return these in a separate dictionary.

    Entries have to share key as well as value to be
    considered common.

    Parameters
    ----------
    A, B : dict
        Dictionaries to merge.
    """

    result = {}

    for key in set(A.keys()) & set(B.keys()):
        v1, v2 = A[key], B[key]
        if type(v1) != type(v2):
            raise TypeError(f"type({v1}) {type(v1)} != type({v2}) {type(v2)}")
        if isinstance(v1, dict) and v1 == v2:
            result[key] = deepcopy(v1)
        elif isinstance(v1, str) and v1 == v2:
            result[key] = A[key]
        elif isinstance(v1, list) and v1 == v2:
            result[key] = deepcopy(v1)
        elif isinstance(v1, np.ndarray) and all(v1 == v2):
            result[key] = deepcopy(v1)
    return result


def _reference_name(class_name):
    """
    Given the name of a class, return an attribute name to be used for
    references to instances of that class.

    For example, a Segment object has a parent Block object, referenced by
    `segment.block`. The attribute name `block` is obtained by calling
    `_container_name("Block")`.
    """
    return class_name.lower()


def _container_name(class_name):
    """
    Given the name of a class, return an attribute name to be used for
    lists (or other containers) containing instances of that class.

    For example, a Block object contains a list of Segment objects,
    referenced by `block.segments`. The attribute name `segments` is
    obtained by calling `_container_name_plural("Segment")`.
    """
    if "RegionOfInterest" in class_name:
        # this is a hack, pending a more principled way to handle this
        return "regionsofinterest"
    else:
        return _reference_name(class_name) + "s"


class BaseNeo:
    """
    This is the base class from which all Neo objects inherit.

    This class implements support for universally recommended arguments,
    and also sets up the :attr:`annotations` dict for additional arguments.

    Each class can define one or more of the following class attributes:
        :_parent_objects: Neo objects that can be parents of this
                          object. Note that no Neo object can have
                          more than one parent.
                          An instance attribute named
                          class.__name__.lower() will be automatically
                          defined to hold this parent and will be
                          initialized to None.
        :_necessary_attrs: A list of tuples containing the attributes that the
                           class must have. The tuple can have 2-4 elements.
                           The first element is the attribute name.
                           The second element is the attribute type.
                           The third element is the number of dimensions
                           (only for numpy arrays and quantities).
                           The fourth element is the dtype of array
                           (only for numpy arrays and quantities).
                           This does NOT include the attributes holding the
                           parents or children of the object.
        :_recommended_attrs: A list of tuples containing the attributes that
                             the class may optionally have.  It uses the same
                             structure as :_necessary_attrs:
        :_repr_pretty_attrs_keys_: The names of attributes printed when
                                   pretty-printing using iPython.

    The following helper properties are available:
        :_parent_containers: The names of the container attributes used
                             to store :_parent_objects:
        :parents: All objects that are parents of the current object.
        :_all_attrs: All required and optional attributes.
                     :_necessary_attrs: + :_recommended_attrs:

    The following "universal" methods are available:
        :__init__: Grabs the universally recommended arguments :attr:`name`,
            :attr:`file_origin`, and :attr:`description` and stores them as
            attributes.

            Also takes every additional argument (that is, every argument
            that is not handled by :class:`BaseNeo` or the child class), and
            puts in the dict :attr:`annotations`.

        :annotate(**args): Updates :attr:`annotations` with keyword/value
                           pairs.

        :merge(**args): Merge the contents of another object into this one.
                        The merge method implemented here only merges
                        annotations (see :merge_annotations:).
                        Subclasses should implementt their own merge rules.

        :merge_annotations(**args): Merge the :attr:`annotations` of another
                                    object into this one.

    Each child class should:
        0) describe its parents (if any) and attributes in the relevant
           class attributes. :_recommended_attrs: should append
           BaseNeo._recommended_attrs to the end.
        1) call BaseNeo.__init__(self, name=name, description=description,
                                 file_origin=file_origin, **annotations)
           with the universal recommended arguments, plus optional annotations
        2) process its required arguments in its __new__ or __init__ method
        3) process its non-universal recommended arguments (in its __new__ or
           __init__ method

    Non-keyword arguments should only be used for required arguments.

    The required and recommended arguments for each child class (Neo object)
    are specified in the _necessary_attrs and _recommended_attrs attributes and
    documentation for the child object.
    """

    # these attributes control relationships, they need to be
    # specified in each child class
    # Parent objects whose children can have a single parent
    _parent_objects = ()
    # Attribute names corresponding to _parent_objects
    _parent_attrs = ()

    # Attributes that an instance is required to have defined
    _necessary_attrs = ()
    # Attributes that an instance may or may have defined
    _recommended_attrs = (("name", str), ("description", str), ("file_origin", str))
    # Attributes that are used for pretty-printing
    _repr_pretty_attrs_keys_ = ("name", "description", "annotations")

    is_view = False

    def __init__(self, name=None, description=None, file_origin=None, **annotations):
        """
        This is the base constructor for all Neo objects.

        Stores universally recommended attributes and creates
        :attr:`annotations` from additional arguments not processed by
        :class:`BaseNeo` or the child class.
        """
        # create `annotations` for additional arguments
        _check_annotations(annotations)
        self.annotations = annotations

        # these attributes are recommended for all objects.
        self.name = name
        self.description = description
        self.file_origin = file_origin

        # initialize parent containers
        for parent in self._parent_containers:
            setattr(self, parent, None)

    def annotate(self, **annotations):
        """
        Add annotations (non-standardized metadata) to a Neo object.

        Example:

        >>> obj.annotate(key1=value0, key2=value1)
        >>> obj.key2
        value2
        """
        _check_annotations(annotations)
        self.annotations.update(annotations)

    def _has_repr_pretty_attrs_(self):
        return any(getattr(self, k) for k in self._repr_pretty_attrs_keys_)

    def _repr_pretty_attrs_(self, pp, cycle):
        first = True
        for key in self._repr_pretty_attrs_keys_:
            value = getattr(self, key)
            if value:
                if first:
                    first = False
                else:
                    pp.breakable()
                with pp.group(indent=1):
                    pp.text(f"{key}: ")
                    pp.pretty(value)

    def _repr_pretty_(self, pp, cycle):
        """
        Handle pretty-printing the :class:`BaseNeo`.
        """
        pp.text(self.__class__.__name__)
        if self._has_repr_pretty_attrs_():
            pp.breakable()
            self._repr_pretty_attrs_(pp, cycle)

    @property
    def _parent_containers(self):
        """
        Containers for parent objects.
        """
        return tuple([_reference_name(parent) for parent in self._parent_objects])

    @property
    def parents(self):
        """
        All parent objects storing the current object.
        """
        return tuple([getattr(self, attr) for attr in self._parent_containers])

    @property
    def _all_attrs(self):
        """
        Returns a combination of all required and recommended
        attributes.
        """
        return self._necessary_attrs + self._recommended_attrs

    def merge_annotations(self, *others):
        """
        Merge annotations from the other object into this one.

        Merging follows these rules:
        All keys that are in the either object, but not both, are kept.
        For keys that are present in both objects:
            For arrays or lists: concatenate the two arrays
            For dicts: merge recursively
            For strings: concatenate with ';'
            Otherwise: fail if the annotations are not equal
        """
        other_annotations = [other.annotations for other in others]
        merged_annotations = merge_annotations(self.annotations, *other_annotations)
        self.annotations.update(merged_annotations)

    def merge(self, *others):
        """
        Merge the contents of another object into this one.

        See :meth:`merge_annotations` for details of the merge operation.
        """
        self.merge_annotations(*others)

    def set_parent(self, obj):
        """
        Set the appropriate "parent" attribute of this object
        according to the type of "obj"
        """
        if obj.__class__.__name__ not in self._parent_objects:
            raise TypeError(
                (
                    f"{self.__class__.__name__} can only have parents of "
                    f"type {self._parwents_objects}, not {obj.__class__.__name__}"
                )
            )
        loc = self._parent_objects.index(obj.__class__.__name__)
        parent_attr = self._parent_attrs[loc]
        setattr(self, parent_attr, obj)



"""
Tests of neo.rawio.spikeglxrawio
"""

import unittest

from neo.rawio.spikeglxrawio import SpikeGLXRawIO
from neo.test.rawiotest.common_rawio_test import BaseTestRawIO
import numpy as np


class TestSpikeGLXRawIO(BaseTestRawIO, unittest.TestCase):
    rawioclass = SpikeGLXRawIO
    entities_to_download = ["spikeglx"]
    entities_to_test = [
        "spikeglx/Noise4Sam_g0",
        "spikeglx/TEST_20210920_0_g0",
        # this is only g0 multi index
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0/5-19-2022-CI0_g0",
        # this is only g1 multi index
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0/5-19-2022-CI0_g1",
        # this mix both multi gate and multi trigger (and also multi probe)
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI0",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI1",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI2",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI3",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI4",
        "spikeglx/multi_trigger_multi_gate/SpikeGLX/5-19-2022-CI5",
        # different sync/sybset options with commercial NP2
        "spikeglx/NP2_with_sync",
        "spikeglx/NP2_no_sync",
        "spikeglx/NP2_subset_with_sync",
        # NP-ultra
        "spikeglx/np_ultra_stub",
        # CatGT
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-A",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-B",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-C",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-D",
        "spikeglx/multi_trigger_multi_gate/CatGT/CatGT-E",
        "spikeglx/multi_trigger_multi_gate/CatGT/Supercat-A",
    ]

    def test_with_location(self):
        rawio = SpikeGLXRawIO(self.get_local_path("spikeglx/Noise4Sam_g0"), load_channel_location=True)
        rawio.parse_header()
        # one of the stream have channel location
        have_location = []
        for sig_anotations in rawio.raw_annotations["blocks"][0]["segments"][0]["signals"]:
            have_location.append("channel_location_0" in sig_anotations["__array_annotations__"])
        assert any(have_location)

    def test_sync(self):
        rawio_with_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_with_sync"), load_sync_channel=True)
        rawio_with_sync.parse_header()
        stream_index = list(rawio_with_sync.header["signal_streams"]["name"]).index("imec0.ap")

        # AP stream has 385 channels
        chunk = rawio_with_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 385

        rawio_no_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_with_sync"), load_sync_channel=False)
        rawio_no_sync.parse_header()

        # AP stream has 384 channels
        chunk = rawio_no_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 384

    def test_no_sync(self):
        # requesting sync channel when there is none raises an error
        with self.assertRaises(ValueError):
            rawio_no_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_no_sync"), load_sync_channel=True)
            rawio_no_sync.parse_header()

    def test_subset_with_sync(self):
        rawio_sub = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_subset_with_sync"), load_sync_channel=True)
        rawio_sub.parse_header()
        stream_index = list(rawio_sub.header["signal_streams"]["name"]).index("imec0.ap")

        # AP stream has 121 channels
        chunk = rawio_sub.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 121

        rawio_sub_no_sync = SpikeGLXRawIO(self.get_local_path("spikeglx/NP2_subset_with_sync"), load_sync_channel=False)
        rawio_sub_no_sync.parse_header()
        # AP stream has 120 channels
        chunk = rawio_sub_no_sync.get_analogsignal_chunk(
            block_index=0, seg_index=0, i_start=0, i_stop=100, stream_index=stream_index
        )
        assert chunk.shape[1] == 120

    def test_nidq_digital_channel(self):
        rawio_digital = SpikeGLXRawIO(self.get_local_path("spikeglx/DigitalChannelTest_g0"))
        rawio_digital.parse_header()
        # This data should have 8 event channels
        assert np.shape(rawio_digital.header["event_channels"])[0] == 8

        # Channel 0 in this data will have sync pulses at 1 Hz, let's confirm that
        all_events = rawio_digital.get_event_timestamps(0, 0, 0)
        on_events = np.where(all_events[2] == "XD0 ON")
        on_ts = all_events[0][on_events]
        on_ts_scaled = rawio_digital.rescale_event_timestamp(on_ts)
        on_diff = np.diff(on_ts_scaled)
        atol = 0.001
        assert np.allclose(on_diff, 1, atol=atol)


if __name__ == "__main__":
    unittest.main()
